"""
Matrix file attachments. Full e2ee support is implemented.
"""

import abc
import enum
import io
import json
import logging
import os
import pathlib
import re
import shutil
import subprocess
import tempfile
import time
import typing
import urllib.parse
import warnings
from typing import Union as U, overload

import PIL.Image
import aiofiles
import aiohttp
import blurhash
import magic
import nio

from .exceptions import (
    MediaCodecWarning,
    MediaDownloadException,
    MediaUploadException,
    MetadataDetectionException,
)
from .utils import run_blocking

if typing.TYPE_CHECKING:
    from .client import NioBot


log = logging.getLogger(__name__)
_CT = typing.TypeVar("_CT", bound=U[str, bytes, pathlib.Path, typing.IO[bytes]])


__all__ = (
    "detect_mime_type",
    "get_metadata_ffmpeg",
    "get_metadata_imagemagick",
    "get_metadata",
    "generate_blur_hash",
    "first_frame",
    "which",
    "BaseAttachment",
    "FileAttachment",
    "ImageAttachment",
    "VideoAttachment",
    "AudioAttachment",
    "AttachmentType",
    "SUPPORTED_CODECS",
    "SUPPORTED_VIDEO_CODECS",
    "SUPPORTED_AUDIO_CODECS",
    "SUPPORTED_IMAGE_CODECS",
)

SUPPORTED_VIDEO_CODECS = [
    "h264",
    "vp8",
    "vp9",
    "av1",
    "theora",
]
# Come on browsers, five codecs is lackluster support. I'm looking at you, Safari.
SUPPORTED_AUDIO_CODECS = [
    "speex",
    "opus",
    "aac",
    "mp3",
    "vorbis",
    "flac",
    "mp2",
]
# All of the above codecs were played in Element Desktop. A bunch were cut out, as the list was far too long.
# Realistically, I don't see the warning being useful to too many people, it's literally only in to help people figure
# out why their media isn't playing.
SUPPORTED_IMAGE_CODECS = ["mjpeg", "gif", "png", "av1", "webp"]
# Probably not all of them but close enough
SUPPORTED_CODECS = SUPPORTED_VIDEO_CODECS + SUPPORTED_AUDIO_CODECS + SUPPORTED_IMAGE_CODECS


def detect_mime_type(file: U[str, io.BytesIO, pathlib.Path]) -> str:
    """
    Detect the mime type of a file.

    :param file: The file to detect the mime type of. Can be a BytesIO.
    :return: The mime type of the file (e.g. `text/plain`, `image/png`, `application/pdf`, `video/webp` etc.)
    """
    if isinstance(file, str):
        file = pathlib.Path(file)

    if isinstance(file, io.BytesIO):
        current_position = file.tell()
        file.seek(0)
        start = time.perf_counter()
        mt = magic.from_buffer(file.read(), mime=True)
        log.debug("Took %f seconds to detect mime type", time.perf_counter() - start)
        file.seek(current_position)  # Reset the file position
        return mt
    elif isinstance(file, pathlib.Path):
        start = time.perf_counter()
        mt = magic.from_file(str(file), mime=True)
        log.debug("Took %f seconds to detect mime type", time.perf_counter() - start)
        return mt
    else:
        raise TypeError("File must be a string, BytesIO, or Path object.")


def get_metadata_ffmpeg(file: U[str, pathlib.Path]) -> dict[str, typing.Any]:
    """
    Gets metadata for a file via ffprobe.

    [example output (JSON)](https://github.com/nexy7574/nio-bot/blob/0e61ea1/docs/assets/guides/text/example_ffprobe.json)

    :param file: The file to get metadata for. **Must be a path-like object**
    :return: A dictionary containing the metadata.
    """
    if not shutil.which("ffprobe"):
        raise FileNotFoundError("ffprobe is not installed. If it is, check your $PATH.")
    command = ["ffprobe", "-of", "json", "-loglevel", "9", "-show_format", "-show_streams", "-i", str(file)]
    start = time.perf_counter()
    try:
        result = subprocess.run(command, capture_output=True, encoding="utf-8", errors="replace", check=True)
    except subprocess.SubprocessError as e:
        raise MetadataDetectionException("Failed to get metadata for file.", exception=e)
    log.debug("Took %f seconds to run ffprobe", time.perf_counter() - start)
    log.debug("ffprobe output (%d): %s", result.returncode, result.stdout)
    data = json.loads(result.stdout or "{}")
    log.debug("parsed ffprobe output:\n%s", json.dumps(data, indent=4))
    return data


def get_metadata_imagemagick(file: pathlib.Path) -> dict[str, typing.Any]:
    """The same as `get_metadata_ffmpeg` but for ImageMagick.

    Only returns a limited subset of the data, such as one stream, which contains the format, and size,
    and the format, which contains the filename, format, and size.

    [example output (JSON)](https://raw.githubusercontent.com/nexy7574/nio-bot/80b33a/docs/assets/guides/text/example_identify.json)

    :param file: The file to get metadata for. **Must be a path object**
    :return: A slimmed-down dictionary containing the metadata.
    """
    file = file.resolve(True)
    command = ["identify", str(file)]
    start = time.perf_counter()
    try:
        result = subprocess.run(command, capture_output=True, encoding="utf-8", errors="replace", check=True)
    except subprocess.SubprocessError as e:
        raise MetadataDetectionException("Failed to get metadata for file.", exception=e)
    log.debug("identify output (%d): %s", result.returncode, result.stdout)
    log.debug("identify took %f seconds", time.perf_counter() - start)
    stdout = result.stdout
    stdout = stdout[len(str(file)) + 1 :]
    img_format, img_size, *_ = stdout.split()
    data = {
        "streams": [
            {
                "index": 0,
                "codec_name": img_format,
                "codec_long_name": img_format,
                "codec_type": "video",
                "height": int(img_size.split("x")[1]),
                "width": int(img_size.split("x")[0]),
            }
        ],
        "format": {
            "filename": str(file),
            "format_long_name": img_format,
            "size": str(file.stat().st_size),
        },
    }
    log.debug("Parsed identify output:\n%s", json.dumps(data, indent=4))
    return data


def get_metadata(file: U[str, pathlib.Path], mime_type: typing.Optional[str] = None) -> dict[str, typing.Any]:
    """
    Gets metadata for a file.

    This will use imagemagick (`identify`) for images where available, falling back to ffmpeg (`ffprobe`)
    for everything else.

    :param file: The file to get metadata for.
    :param mime_type: The mime type of the file. If not provided, it will be detected.
    :return: The metadata for the file. See [niobot.get_metadata_ffmpeg][] and [niobot.get_metadata_imagemagick][]
     for more information.
    """
    file = _to_path(file)
    mime = mime_type or detect_mime_type(file)
    mime = mime.split("/")[0]
    if mime == "image":
        if not shutil.which("identify"):
            log.warning(
                "Imagemagick identify not found, falling back to ffmpeg for image metadata detection. "
                "Check your $PATH."
            )
        else:
            start = time.perf_counter()

            try:
                r = get_metadata_imagemagick(file)
                log.debug("get_metadata_imagemagick took %f seconds", time.perf_counter() - start)
                return r
            except (IndexError, ValueError, subprocess.SubprocessError, IOError, OSError):
                log.warning(
                    "Failed to detect metadata for %r with imagemagick. Falling back to ffmpeg.", file, exc_info=True
                )

    if mime not in ["audio", "video", "image"]:
        raise MetadataDetectionException("Unsupported mime type. Must be an audio clip, video, or image.")
    start = time.perf_counter()
    r = get_metadata_ffmpeg(file)
    log.debug("get_metadata_ffmpeg took %f seconds", time.perf_counter() - start)
    return r


def first_frame(file: U[str, pathlib.Path], file_format: str = "webp") -> bytes:
    """
    Gets the first frame of a video file.

    !!! Danger "This function creates a file on disk"
        In order to extract the frame, this function creates a temporary file on disk (or memdisk depending on where
        your tempdir is). While this file is deleted after the function is done, it is still something to be aware of.
        For example, if you're (worryingly) low on space, this function may fail to extract the frame due to a lack of
        space. Or, someone could temporarily access and read the file before it is deleted.

        This also means that this function may be slow.

    :param file: The file to get the first frame of. **Must be a path-like object**
    :param file_format: The format to save the frame as. Defaults to webp.
    :return: The first frame of the video in bytes.
    """
    if not shutil.which("ffmpeg"):
        raise FileNotFoundError("ffmpeg is not installed. If it is, check your $PATH.")
    with tempfile.NamedTemporaryFile(suffix=f".{file_format}") as f:
        command = ["ffmpeg", "-loglevel", "9", "-i", str(file), "-frames:v", "1", "-y", "-strict", "-2", f.name]
        log.debug("Extracting first frame of %r: %s", file, " ".join(command))
        try:
            start = time.perf_counter()
            log.debug("Extraction return code: %d", subprocess.run(command, capture_output=True, check=True).returncode)
            log.debug("Extraction took %f seconds", time.perf_counter() - start)
        except subprocess.SubprocessError as e:
            raise MediaUploadException("Failed to extract first frame of video.", exception=e)
        f.seek(0)
        return f.read()


def generate_blur_hash(file: U[str, pathlib.Path, io.BytesIO, PIL.Image.Image], *parts: int) -> str:
    """
    Creates a blurhash

    !!! warning "This function may be resource intensive"
        This function may be resource intensive, especially for large images. You should run this in a thread or
        process pool.

        You should also scale any images down in order to increase performance.

        See: [woltapp/blurhash](https://github.com/woltapp/blurhash)
    """
    if not parts:
        parts = 4, 3
    if not isinstance(file, (io.BytesIO, PIL.Image.Image)):
        file = _to_path(file)
        size = PIL.Image.open(file).size
        with file.open("rb") as fd:
            log.info("Generating blurhash for %s (%s)", file, size)
            start = time.perf_counter()
            x = blurhash.encode(fd, *parts)
            log.debug("Generating blurhash took %f seconds", time.perf_counter() - start)
            return x
    else:
        if isinstance(file, io.BytesIO):
            size = PIL.Image.open(file).size
        else:
            size = file.size
        log.info("Generating blurhash for BytesIO/PIL object %r (%s)", file, size)
        start = time.perf_counter()
        x = blurhash.encode(file, *parts)
        log.debug("Generating blurhash took %f seconds", time.perf_counter() - start)
        return x


def _file_okay(file: U[pathlib.Path, io.BytesIO]) -> typing.Literal[True]:
    """Checks if a file exists, is a file, and can be read."""
    if isinstance(file, io.BytesIO):
        if file.closed:
            raise ValueError("BytesIO object is closed.")
        if len(file.getbuffer()) == 0:
            w = ResourceWarning("BytesIO object is empty, this may cause issues. Did you mean to seek(0) first?")
            warnings.warn(w)
        return True

    if not file.exists():
        raise FileNotFoundError(f"File {file} does not exist.")
    if not file.is_file():
        raise ValueError(f"{file} is not a file.")
    if not os.access(file, os.R_OK):
        raise PermissionError(f"Cannot read {file}.")
    # Check it can have a stat() value
    file.stat()
    return True


@overload
def _to_path(file: U[str, pathlib.Path]) -> pathlib.Path: ...


@overload
def _to_path(file: io.BytesIO) -> io.BytesIO: ...


def _to_path(file: U[str, pathlib.Path, io.BytesIO]) -> U[pathlib.Path, io.BytesIO]:
    """Converts a string to a Path object."""
    if not isinstance(file, (str, pathlib.Path, io.BytesIO)):
        raise TypeError("File must be a string, BytesIO, or Path object.")

    if isinstance(file, io.BytesIO):
        return file

    if isinstance(file, str):
        file = pathlib.Path(file)
    file = file.resolve()
    return file


def _size(file: U[pathlib.Path, io.BytesIO]) -> int:
    """Gets the size of a file."""
    if isinstance(file, io.BytesIO):
        return len(file.getbuffer())
    return file.stat().st_size


def which(
    file: U[io.BytesIO, pathlib.Path, str], mime_type: typing.Optional[str] = None
) -> U[
    typing.Type["FileAttachment"],
    typing.Type["ImageAttachment"],
    typing.Type["AudioAttachment"],
    typing.Type["VideoAttachment"],
]:
    """
    Gets the correct attachment type for a file.

    This function will provide either Image/Video/Audio attachment where possible, or FileAttachment otherwise.

    For example, `image/png` (from `my_image.png`) will see `image/` and will return
    [`ImageAttachment`][niobot.ImageAttachment], and `video/mp4` (from `my_video.mp4`) will see `video/` and will
    return [`VideoAttachment`][niobot.VideoAttachment].

    If the mime type cannot be mapped to an attachment type, this function will return
    [`FileAttachment`][niobot.FileAttachment].

    ??? example "Usage"
        ```python
        import niobot
        import pathlib

        my_file = pathlib.Path("/tmp/foo.bar")
        attachment = await niobot.which(my_file).from_file(my_file)
        # or
        attachment_type = niobot.which(my_file)  # one of the BaseAttachment subclasses
        attachment = await attachment_type.from_file(my_file)
        ```

    :param file: The file or BytesIO to investigate
    :param mime_type: The optional pre-detected mime type. If this is not provided, it will be detected.
    :return: The correct type for this attachment (not instantiated)
    """
    values = {
        "image": ImageAttachment,
        "audio": AudioAttachment,
        "video": VideoAttachment,
    }
    if not mime_type:
        mime_type = detect_mime_type(file)
    mime_start = mime_type.split("/")[0].lower()
    return values.get(mime_start, FileAttachment)


class AttachmentType(enum.Enum):
    """
    Enumeration containing the different types of media.

    :var FILE: A generic file.
    :var AUDIO: An audio file.
    :var VIDEO: A video file.
    :var IMAGE: An image file.
    """

    if typing.TYPE_CHECKING:
        FILE: "AttachmentType"
        AUDIO: "AttachmentType"
        VIDEO: "AttachmentType"
        IMAGE: "AttachmentType"
    FILE = "m.file"
    AUDIO = "m.audio"
    VIDEO = "m.video"
    IMAGE = "m.image"


class BaseAttachment(abc.ABC):
    """
    Base class for attachments

    !!! note
        If you pass a custom `file_name`, this is only actually used if you pass a [io.BytesIO][] to `file`.
        If you pass a [pathlib.Path][] or a [string][str], the file name will be resolved from the path, overriding
        the `file_name` parameter.

    :param file: The file path or BytesIO object to upload.
    :param file_name: The name of the file. **Must be specified if uploading a BytesIO object.**
    :param mime_type: The mime type of the file. If not specified, it will be detected.
    :param size_bytes: The size of the file in bytes. If not specified, it will be detected.
    :param attachment_type: The type of attachment. Defaults to `AttachmentType.FILE`.

    :ivar file: The file path or BytesIO object to upload. Resolved to a [pathlib.Path][] object if a string is
    passed to `__init__`.
    :ivar file_name: The name of the file. If `file` was a string or `Path`, this will be the name of the file.
    :ivar mime_type: The mime type of the file.
    :ivar size: The size of the file in bytes.
    :ivar type: The type of attachment.
    :ivar url: The URL of the uploaded file. This is set after the file is uploaded.
    :ivar keys: The encryption keys for the file. This is set after the file is uploaded.
    """

    if typing.TYPE_CHECKING:
        file: U[pathlib.Path, io.BytesIO]
        file_name: str
        mime_type: str
        size: int
        type: AttachmentType

        url: typing.Optional[str]
        keys: typing.Optional[dict[str, str]]

    def __init__(
        self,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        size_bytes: typing.Optional[int] = None,
        *,
        attachment_type: AttachmentType = AttachmentType.FILE,
    ):
        self.file = _to_path(file)
        # Ignore type error as the type is checked right afterwards
        self.file_name = self.file.name if isinstance(self.file, pathlib.Path) else file_name  # type: ignore
        if not self.file_name:
            raise ValueError("file_name must be specified when uploading a BytesIO object.")
        self.mime_type = mime_type or detect_mime_type(self.file)
        if size_bytes:
            self.size = size_bytes
        elif isinstance(self.file, io.BytesIO):
            self.size = len(self.file.getbuffer())
        else:
            os.path.getsize(self.file)

        self.type = attachment_type
        self.url = None
        self.keys = None

    def __repr__(self):
        return (
            "<{0.__class__.__name__} file={0.file!r} file_name={0.file_name!r} "
            "mime_type={0.mime_type!r} size={0.size!r} type={0.type!r}>".format(self)
        )

    def as_body(self, body: typing.Optional[str] = None) -> dict:
        """
        Generates the body for the attachment for sending. The attachment must've been uploaded first.

        :param body: The body to use (should be a textual description). Defaults to the file name.
        :return:
        """
        output_body = {
            "body": body or self.file_name,
            "info": {
                "mimetype": self.mime_type,
                "size": self.size,
            },
            "msgtype": self.type.value,
            "filename": self.file_name,
            "url": self.url,
        }
        if self.keys:
            output_body["file"] = self.keys
        return output_body

    @classmethod
    async def from_file(
        cls,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
    ) -> "BaseAttachment":
        """
        Creates an attachment from a file.

        You should use this method instead of the constructor, as it will automatically detect all other values

        :param file: The file or BytesIO to attach
        :param file_name: The name of the BytesIO file, if applicable
        :return: Loaded attachment.
        """
        file = _to_path(file)
        if not file_name:
            if isinstance(file, io.BytesIO):
                raise ValueError("file_name must be specified when uploading a BytesIO object.")
            else:
                file_name = file.name

        mime_type = await run_blocking(detect_mime_type, file)
        size = _size(file)
        return cls(file, file_name, mime_type, size)

    @classmethod
    async def from_mxc(cls, client: "NioBot", url: str) -> "BaseAttachment":
        """
        Creates an attachment from an MXC URL.

        :param client: The current client instance (used to download the attachment)
        :param url: The MXC:// url to download
        :return: The downloaded and probed attachment.
        """
        response = await client.download(url)
        if isinstance(response, nio.DownloadResponse):
            return await cls.from_file(io.BytesIO(response.body), response.filename)
        raise MediaDownloadException("Failed to download attachment.", response)

    @classmethod
    async def from_http(
        cls,
        url: str,
        client_session: typing.Optional[aiohttp.ClientSession] = None,
    ) -> "BaseAttachment":
        """
        Creates an attachment from an HTTP URL.

        This is not necessarily just for images, video, or other media - it can be used for any HTTP resource.

        :param url: The http/s URL to download
        :param client_session: The aiohttp client session to use. If not specified, a new one will be created.
        :return: The downloaded and probed attachment.
        :raises niobot.MediaDownloadException: if the download failed.
        :raises aiohttp.ClientError: if the download failed.
        :raises niobot.MediaDetectionException: if the MIME type could not be detected.
        """
        if not client_session:
            from . import __user_agent__

            async with aiohttp.ClientSession(headers={"User-Agent": __user_agent__}) as session:
                return await cls.from_http(url, session)

        async with client_session.get(url) as response:
            try:
                response.raise_for_status()
            except aiohttp.ClientResponseError as err:
                raise MediaDownloadException("Failed to download attachment.", exception=err)

            file_name = response.headers.get("Content-Disposition")
            if file_name:
                file_name = re.search(r"filename=\"(.+)\"", file_name)
                if file_name:
                    file_name = file_name.group(1)

            if not file_name:
                u = urllib.parse.urlparse(url)
                file_name = os.path.basename(u.path)

            return await cls.from_file(io.BytesIO(await response.read()), file_name)

    @property
    def size_bytes(self) -> int:
        """Returns the size of this attachment in bytes."""
        return self.size

    def size_as(
        self,
        unit: typing.Literal[
            "b",
            "kb",
            "kib",
            "mb",
            "mib",
            "gb",
            "gib",
        ],
    ) -> U[int, float]:
        """
        Helper function to convert the size of this attachment into a different unit.

        Remember:
        - 1 kilobyte (KB) is 1000 bytes
        - 1 kibibyte (KiB) is 1024 bytes

        ??? example "Example"
            ```python
            >>> import niobot
            >>> attachment = niobot.FileAttachment("background.png", "image/png")
            >>> attachment.size_bytes
            329945
            >>> attachment.size_as("kb")
            329.945
            >>> attachment.size_as("kib")
            322.2119140625
            >>> attachment.size_as("mb")
            0.329945
            >>> attachment.size_as("mib")
            0.31466007232666016
            ```
            *Note that due to the nature of floats, precision may be lost, especially the larger in units you go.*

        :param unit: The unit to convert into
        :return: The converted size
        """
        multi = {
            "b": 1,
            "kb": 1000,
            "kib": 1024,
            "mb": 1000**2,
            "mib": 1024**2,
            "gb": 1000**3,
            "gib": 1024**3,
        }
        return self.size_bytes / multi[unit]

    async def upload(self, client: "NioBot", encrypted: bool = False) -> "BaseAttachment":
        """
        Uploads the file to matrix.

        :param client: The client to upload
        :param encrypted: Whether to encrypt the attachment or not
        :return: The attachment
        """
        if self.keys or self.url:
            raise RuntimeError("This attachment has already been uploaded.")
        if self.file_name is None:
            if hasattr(self.file, "name"):
                self.file_name = self.file.name
            else:
                raise ValueError("file_name must be specified when uploading a BytesIO object.")
        size = self.size or _size(self.file)

        if not isinstance(self.file, io.BytesIO):
            # We can open the file async here, as this will avoid blocking the loop
            async with aiofiles.open(self.file, "rb") as f:
                log.debug("Using aiofiles to open %r" % self.file)
                result, keys = await client.upload(
                    f,
                    content_type=self.mime_type,
                    filename=self.file_name,
                    encrypt=encrypted,
                    filesize=size,
                )
        else:
            # Usually, BytesIO objects are small enough to be uploaded synchronously. Plus, they're literally just
            # in-memory.
            # For scale, here is a 1GiB BytesIO with urandom() content, seek(0)ed and read() in its entirety, with
            # timeit:
            # 47.2 ns ± 0.367 ns per loop (mean ± std. dev. of 7 runs, 10,000,000 loops each)
            # So in reality, it's not going to be a massive problem.
            pos = self.file.tell()
            self.file.seek(0)
            log.debug("Uploading BytesIO object, seeked from position %d to 0, will restore after upload." % pos)
            result, keys = await client.upload(
                self.file,
                content_type=self.mime_type,
                filename=self.file_name,
                encrypt=encrypted,
                filesize=size,
            )
            self.file.seek(pos)
            log.debug("Uploaded BytesIO, seeked back to position %d." % pos)
        if not isinstance(result, nio.UploadResponse):
            raise MediaUploadException("Upload failed: %r" % result, result)

        if keys:
            self.keys = keys

        self.url = result.content_uri
        return self


class FileAttachment(BaseAttachment):
    """
    Represents a generic file attachment.

    You should use [VideoAttachment][niobot.attachment.VideoAttachment] for videos,
    [AudioAttachment][niobot.attachment.AudioAttachment] for audio,
    and [ImageAttachment][niobot.attachment.ImageAttachment] for images.
    This is for everything else.

    :param file: The file to upload
    :param file_name: The name of the file
    :param mime_type: The mime type of the file
    :param size_bytes: The size of the file in bytes
    """

    def __init__(
        self,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        size_bytes: typing.Optional[int] = None,
    ):
        super().__init__(file, file_name, mime_type, size_bytes, attachment_type=AttachmentType.FILE)


class ImageAttachment(BaseAttachment):
    """
    Represents an image attachment.

    :param file: The file to upload
    :param file_name: The name of the file
    :param mime_type: The mime type of the file
    :param size_bytes: The size of the file in bytes
    :param height: The height of the image in pixels (e.g. 1080)
    :param width: The width of the image in pixels (e.g. 1920)
    :param thumbnail: A thumbnail of the image. NOT a blurhash.
    :param xyz_amorgan_blurhash: The blurhash of the image

    :ivar info: A dict of info about the image. Contains `h`, `w`, `mimetype`, and `size` keys.
    :ivar thumbnail: A thumbnail of the image. NOT a blurhash.
    """

    def __init__(
        self,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        size_bytes: typing.Optional[int] = None,
        height: typing.Optional[int] = None,
        width: typing.Optional[int] = None,
        thumbnail: typing.Optional["ImageAttachment"] = None,
        xyz_amorgan_blurhash: typing.Optional[str] = None,
    ):
        super().__init__(
            file,
            file_name,
            mime_type,
            size_bytes,
            attachment_type=AttachmentType.IMAGE,
        )
        self.xyz_amorgan_blurhash = xyz_amorgan_blurhash
        self.info = {
            "h": height,
            "w": width,
            "mimetype": mime_type,
            "size": size_bytes,
        }
        self.thumbnail = thumbnail

    @property
    def height(self) -> typing.Optional[int]:
        """The height of this image in pixels"""
        return self.info["h"]

    @height.setter
    def height(self, value: typing.Optional[int]):
        self.info["h"] = value

    @property
    def width(self) -> typing.Optional[int]:
        """The width of this image in pixels"""
        return self.info["w"]

    @width.setter
    def width(self, value: typing.Optional[int]):
        self.info["w"] = value

    @classmethod
    async def from_file(
        cls,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        height: typing.Optional[int] = None,
        width: typing.Optional[int] = None,
        thumbnail: typing.Optional["ImageAttachment"] = None,
        generate_blurhash: bool = True,
        *,
        xyz_amorgan_blurhash: typing.Optional[str] = None,
        unsafe: bool = False,
    ) -> "ImageAttachment":
        """
        Generates an image attachment

        :param file: The file to upload
        :param file_name: The name of the file (only used if file is a `BytesIO`)
        :param height: The height, in pixels, of this image
        :param width: The width, in pixels, of this image
        :param thumbnail: A thumbnail for this image
        :param generate_blurhash: Whether to generate a blurhash for this image
        :param xyz_amorgan_blurhash: The blurhash of the image, if known beforehand.
        :param unsafe: Whether to allow uploading of images with unsupported codecs. May break metadata detection.
        :return: An image attachment
        """
        file = _to_path(file)
        if isinstance(file, io.BytesIO):
            if not file_name:
                raise ValueError("file_name must be specified when uploading a BytesIO object.")
        else:
            if not file_name:
                file_name = file.name

            if height is None or width is None:
                metadata = await run_blocking(get_metadata, file)
                for stream in metadata["streams"]:
                    log.debug("Found stream in image:\n%s", stream)
                    if stream["codec_type"] == "video":
                        if stream["codec_name"].lower() not in SUPPORTED_IMAGE_CODECS and unsafe is False:
                            warning = MediaCodecWarning(stream["codec_name"], *SUPPORTED_IMAGE_CODECS)
                            warnings.warn(warning)
                        log.debug("Selecting stream %r for image", stream)
                        break
                else:
                    raise ValueError("Unable to find an image stream in the given file. Are you sure its an image?")
                # ffmpeg doesn't have an image type
                height = stream["height"]
                width = stream["width"]

        mime_type = await run_blocking(detect_mime_type, file)
        size = _size(file)
        self = cls(
            file, file_name, mime_type, size, height, width, thumbnail, xyz_amorgan_blurhash=xyz_amorgan_blurhash
        )
        if generate_blurhash:
            await self.get_blurhash()
        return self

    def as_body(self, body: typing.Optional[str] = None) -> dict:
        output_body = super().as_body(body)
        info = self.info.copy()

        # Remove null elements for width and height
        if info.get("h") is None:
            info.pop("h", None)
        if info.get("w") is None:
            info.pop("w", None)
        output_body["info"] = {**output_body["info"], **info}
        if self.thumbnail:
            if self.thumbnail.keys:
                output_body["info"]["thumbnail_file"] = self.thumbnail.keys
            output_body["info"]["thumbnail_info"] = self.thumbnail.info
            output_body["info"]["thumbnail_url"] = self.thumbnail.url
        if self.xyz_amorgan_blurhash:
            output_body["info"]["xyz.amorgan.blurhash"] = self.xyz_amorgan_blurhash
        return output_body

    @staticmethod
    def thumbnailify_image(
        image: U[PIL.Image.Image, io.BytesIO, str, pathlib.Path],
        size: typing.Tuple[int, int] = (320, 240),
        resampling: typing.Union["PIL.Image.Resampling"] = PIL.Image.Resampling.BICUBIC,
    ) -> PIL.Image.Image:
        """
        Helper function to thumbnail an image.

        This function is blocking - you should use [niobot.utils.run_blocking][] to run it.

        :param image: The image to thumbnail
        :param size: The size to thumbnail to. Defaults to 320x240, a standard thumbnail size.
        :param resampling: The resampling filter to use. Defaults to `PIL.Image.BICUBIC`, a high-quality but
        fast resampling method. For the highest quality, use `PIL.Image.LANCZOS`.
        :return: The thumbnail
        """
        if not isinstance(image, PIL.Image.Image):
            image = _to_path(image)
            image = PIL.Image.open(image)
        image.thumbnail(size, resampling)
        return image

    async def get_blurhash(
        self,
        quality: typing.Tuple[int, int] = (4, 3),
        file: typing.Optional[U[str, pathlib.Path, io.BytesIO, PIL.Image.Image]] = None,
        disable_auto_crop: bool = False,
    ) -> str:
        """
        Gets the blurhash of the attachment. See: [woltapp/blurhash](https://github.com/woltapp/blurhash)

        !!! tip "You should crop-down your blurhash images."
            Generating blurhashes can take a long time, *especially* on large images.
            You should crop-down your images to a reasonable size before generating the blurhash.

            Remember, most image quality is lost - there's very little point in generating a blurhash for a 4K image.
            Anything over 800x600 is definitely overkill.

            You can easily resize images with
            [niobot.ImageAttachment.thumbnailify_image][]:

            ```python
            attachment = await niobot.ImageAttachment.from_file(my_image, generate_blurhash=False)
            await attachment.get_blurhash(file=attachment.thumbnailify_image(attachment.file))
            ```

            This will generate a roughly 320x240 thumbnail image, and generate the blurhash from that.

            !!! tip "New!"
                Unless you pass `disable_auto_crop=True`, this function will automatically crop the image down to
                a reasonable size, before generating a blurhash.


        :param quality: A tuple of the quality to generate the blurhash at. Defaults to (4, 3).
        :param file: The file to generate the blurhash from. Defaults to the file passed in the constructor.
        :param disable_auto_crop: Whether to disable automatic cropping of the image. Defaults to False.
        :return: The blurhash
        """
        if isinstance(self.xyz_amorgan_blurhash, str):
            return self.xyz_amorgan_blurhash

        file = file or self.file
        if not isinstance(file, PIL.Image.Image):
            file = _to_path(file)
            file = PIL.Image.open(file)

        if disable_auto_crop is False and (file.width > 800 or file.height > 600):
            log.debug("Cropping image down from {0.width}x{0.height} to 800x600 for faster blurhashing".format(file))
            file.thumbnail((800, 600), PIL.Image.Resampling.BICUBIC)
        x = await run_blocking(generate_blur_hash, file or self.file, *quality)
        self.xyz_amorgan_blurhash = x
        return x


class VideoAttachment(BaseAttachment):
    """
    Represents a video attachment.

    :param file: The file to upload
    :param file_name: The name of the file
    :param mime_type: The mime type of the file
    :param size_bytes: The size of the file in bytes
    :param height: The height of the video in pixels (e.g. 1080)
    :param width: The width of the video in pixels (e.g. 1920)
    :param duration: The duration of the video in seconds
    :param thumbnail: A thumbnail of the video. NOT a blurhash.
    """

    def __init__(
        self,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        size_bytes: typing.Optional[int] = None,
        duration: typing.Optional[int] = None,
        height: typing.Optional[int] = None,
        width: typing.Optional[int] = None,
        thumbnail: typing.Optional["ImageAttachment"] = None,
    ):
        super().__init__(file, file_name, mime_type, size_bytes, attachment_type=AttachmentType.VIDEO)
        self.info = {
            "duration": duration,
            "h": height,
            "w": width,
            "mimetype": mime_type,
            "size": size_bytes,
        }
        self.thumbnail = thumbnail

    @property
    def duration(self) -> typing.Optional[int]:
        """The duration of this video in milliseconds"""
        return self.info["duration"]

    @duration.setter
    def duration(self, value: typing.Optional[int]):
        self.info["duration"] = value

    @property
    def height(self) -> typing.Optional[int]:
        """The height of this image in pixels"""
        return self.info["h"]

    @height.setter
    def height(self, value: typing.Optional[int]):
        self.info["h"] = value

    @property
    def width(self) -> typing.Optional[int]:
        """The width of this image in pixels"""
        return self.info["w"]

    @width.setter
    def width(self, value: typing.Optional[int]):
        self.info["w"] = value

    @classmethod
    async def from_file(
        cls,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        duration: typing.Optional[int] = None,
        height: typing.Optional[int] = None,
        width: typing.Optional[int] = None,
        thumbnail: typing.Optional[U[ImageAttachment, typing.Literal[False]]] = None,
        generate_blurhash: bool = True,
    ) -> "VideoAttachment":
        """
        Generates a video attachment

        !!! warning "This function auto-generates a thumbnail!"
            As thumbnails greatly improve user experience, even with blurhashes enabled, this function will by default
            create a thumbnail of the first frame of the given video if you do not provide one yourself.
            **This may increase your initialisation time by a couple seconds, give or take!**

            If this is undesirable, pass `thumbnail=False` to disable generating a thumbnail.
            This is independent of `generate_blurhash`.

            Generated thumbnails are always WebP images, so they will always be miniature, so you shouldn't
            notice a significant increase in upload time, especially considering your video will likely be several
            megabytes.

        :param file: The file to upload
        :param file_name: The name of the file (only used if file is a `BytesIO`)
        :param duration: The duration of the video, in seconds
        :param height: The height, in pixels, of this video
        :param width: The width, in pixels, of this video
        :param thumbnail: A thumbnail for this image
        :param generate_blurhash: Whether to generate a blurhash for this image
        :return: An image attachment
        """
        file = _to_path(file)
        if isinstance(file, io.BytesIO):
            if not file_name:
                raise ValueError("file_name must be specified when uploading a BytesIO object.")
        else:
            if not file_name:
                file_name = file.name

            if height is None or width is None or duration is None:
                metadata = await run_blocking(get_metadata, file)
                for stream in metadata["streams"]:
                    if stream["codec_type"] == "video":
                        if stream["codec_name"].lower() not in SUPPORTED_VIDEO_CODECS or not stream[
                            "codec_name"
                        ].startswith("pcm_"):  # usually, pcm is supported.
                            warning = MediaCodecWarning(stream["codec_name"], *SUPPORTED_VIDEO_CODECS)
                            warnings.warn(warning)
                        height = stream["height"]
                        width = stream["width"]
                        duration = round(float(metadata["format"]["duration"]) * 1000)
                        break
                else:
                    raise ValueError("Could not find a video stream in this file.")

        mime_type = await run_blocking(detect_mime_type, file)
        size = _size(file)
        original_thumbnail = thumbnail
        if thumbnail is False:
            thumbnail = None
        # noinspection PyTypeChecker
        self = cls(file, file_name, mime_type, size, duration, height, width, thumbnail)
        if generate_blurhash:
            if isinstance(self.thumbnail, ImageAttachment):
                await self.thumbnail.get_blurhash()
            elif isinstance(file, pathlib.Path) and original_thumbnail is not False:
                thumbnail_bytes = await run_blocking(first_frame, file)
                self.thumbnail = await ImageAttachment.from_file(
                    io.BytesIO(thumbnail_bytes), file_name="thumbnail.webp"
                )
        return self

    @staticmethod
    async def generate_thumbnail(video: U[str, pathlib.Path, "VideoAttachment"]) -> ImageAttachment:
        """
        Generates a thumbnail for a video.

        :param video: The video to generate a thumbnail for
        :return: The path to the generated thumbnail
        """
        if isinstance(video, VideoAttachment):
            if not isinstance(video.file, pathlib.Path):
                raise ValueError(
                    "VideoAttachment.file must be a pathlib.Path, BytesIOs are not supported for thumbnail generation"
                )
            video = video.file
        video = _to_path(video)
        x = await run_blocking(first_frame, video, "webp")
        return await ImageAttachment.from_file(io.BytesIO(x), file_name="thumbnail.webp")

    def as_body(self, body: typing.Optional[str] = None) -> dict:
        output_body = super().as_body(body)
        info = self.info.copy()
        # Remove null elements for width, height, and duration
        if info.get("h") is None:
            info.pop("h", None)
        if info.get("w") is None:
            info.pop("w", None)
        if info.get("duration") is None:
            info.pop("duration", None)
        output_body["info"] = {**output_body["info"], **info}
        if self.thumbnail:
            if self.thumbnail.keys:
                output_body["info"]["thumbnail_file"] = self.thumbnail.keys
            output_body["info"]["thumbnail_info"] = self.thumbnail.info
            output_body["info"]["thumbnail_url"] = self.thumbnail.url
        return output_body


class AudioAttachment(BaseAttachment):
    """
    Represents an audio attachment.
    """

    def __init__(
        self,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        mime_type: typing.Optional[str] = None,
        size_bytes: typing.Optional[int] = None,
        duration: typing.Optional[int] = None,
    ):
        super().__init__(file, file_name, mime_type, size_bytes, attachment_type=AttachmentType.AUDIO)
        self.info = {
            "duration": duration,
            "mimetype": mime_type,
            "size": size_bytes,
        }

    @property
    def duration(self) -> typing.Optional[int]:
        """The duration of this audio in milliseconds"""
        return self.info["duration"]

    @duration.setter
    def duration(self, value: typing.Optional[int]):
        self.info["duration"] = value

    @classmethod
    async def from_file(
        cls,
        file: U[str, io.BytesIO, pathlib.Path],
        file_name: typing.Optional[str] = None,
        duration: typing.Optional[int] = None,
    ) -> "AudioAttachment":
        """
        Generates an audio attachment

        :param file: The file to upload
        :param file_name: The name of the file (only used if file is a `BytesIO`)
        :param duration: The duration of the audio, in seconds
        :return: An audio attachment
        """
        file = _to_path(file)
        if isinstance(file, io.BytesIO):
            if not file_name:
                raise ValueError("file_name must be specified when uploading a BytesIO object.")
        else:
            if not file_name:
                file_name = file.name
            if duration is None:
                metadata = await run_blocking(get_metadata, file)
                duration = round(float(metadata["format"]["duration"]) * 1000)

        mime_type = await run_blocking(detect_mime_type, file)
        size = _size(file)
        self = cls(file, file_name, mime_type, size, duration)
        return self

    def as_body(self, body: typing.Optional[str] = None) -> dict:
        output_body = super().as_body(body)
        info = self.info.copy()
        if info.get("duration") is None:
            info.pop("duration", None)
        output_body["info"] = {**output_body["info"], **self.info}
        return output_body



import asyncio
import logging
import typing

if typing.TYPE_CHECKING:
    from ..client import NioBot


__all__ = ("Typing",)

log = logging.getLogger(__name__)
_TYPING_STATES: dict[str, "Typing"] = {}


class Typing:
    """
    Context manager to manage typing notifications.

    :param client: The `NioBot` instance
    :param room_id: The room id to send the typing notification to
    :param timeout: The timeout in seconds
    :param persistent: Whether to send a typing notification every `timeout` seconds, to keep the typing status active
    """

    def __init__(self, client: "NioBot", room_id: str, *, timeout: int = 30, persistent: bool = True):
        if room_id in _TYPING_STATES:
            log.warning("A typing notification is already active for this room: %s", _TYPING_STATES[room_id])
        self.room_id = room_id
        self.client = client
        self.persistent = persistent
        self._task = None
        self.timeout = timeout * 1000

    async def persistence_loop(self):
        while True:
            _TYPING_STATES[self.room_id] = self
            await self.client.room_typing(self.room_id, True, timeout=self.timeout)
            await asyncio.sleep(self.timeout - 1000)
            _TYPING_STATES[self.room_id] = self

    async def __aenter__(self):
        """Starts the typing notification loop, or sends a single typing notification if not persistent."""
        if self.room_id in _TYPING_STATES:
            raise RuntimeError("A typing notification is already active for this room", _TYPING_STATES[self.room_id])
        if not self.persistent:
            await self.client.room_typing(self.room_id, True, timeout=self.timeout)
        else:
            self._task = asyncio.create_task(self.persistence_loop())

    async def __aexit__(self, exc_type, exc, tb):
        """Cancels any existing typing loop under this instance and sends a typing notification to stop typing."""
        if self._task:
            self._task.cancel()
        await self.client.room_typing(self.room_id, False)
        _TYPING_STATES.pop(self.room_id, None)



from .checks import *
from .federation import *
from .help_command import *
from .lib import *
from .mentions import Mentions as Mentions
from .parsers import *
from .string_view import *
from .typing import *
from .unblocking import *



# THIS FILE IS NOT MEANT TO BE IMPORTED! IT DOES NOT FEATURE ANY ACTUAL CODE! ITS JUST STUBS!!!
# This file is written so that mkdocs can auto-generate an event reference.
import typing as t

if t.TYPE_CHECKING:
    from . import CommandError, Context, MatrixRoom, RoomMessage, SyncResponse


async def event_loop_running() -> t.Optional[t.Any]:
    """
    An event that is fired once the event loop is running.

    !!! tip "You should use this event to perform any startup tasks."
        This event is fired before the bot logs in, and before the first `sync()` is performed.

        This means that if, for example, you wanted to initialise a database, or make some HTTP request in a module,
        You can @[nio]bot.event("event_loop_running") do it here.

        ??? example "Initialising a database in a module"
            ```python
            import niobot
            import aiosqlite

            class MyModule(niobot.Module):
                def __init__(self, bot):
                    super().__init__(bot)
                    self.db = None

                @niobot.event("event_loop_running")
                async def event_loop_running(self):
                    self.db = await aiosqlite.connect("mydb.db")
                    await self.db.execute(...)
                    await self.db.commit
            ```
    """
    ...


async def ready(result: "SyncResponse") -> t.Optional[t.Any]:
    """
    An event that is fired when the bot's first `sync()` is completed.

    This indicates that the bot successfully logged in, synchronised with the server, and is ready to receive events.

    :param result: The response from the sync.
    """
    ...


async def message(room: "MatrixRoom", event: "RoomMessage") -> t.Optional[t.Any]:
    """
    An event that is fired when the bot receives a message in a room that it is in.

    This event is dispatched *before* commands are processed, and as such the convenient [niobot.Context][] is
    unavailable.

    !!! danger "Not every message is a text message"
        As of v1.2.0, the `message` event is dispatched for every decrypted message type, as such
        including videos, images, audio, and text. Prior for v1.2.0, this was only dispatched for
        text messages.

        Please check either the type of the event (i.e. `isinstance(event, niobot.RoomMessageText)`)
        or the `event.source["content"]["msgtype"]` to determine the type of the message.

    !!! tip
        If you want to be able to use the [niobot.Context][] in your event handlers, you should use the
        `command` event instead.

        Furthermore, if you want more fine-grained control over how commands are parsed and handled, you should
        *override* [niobot.NioBot.process_message][] instead of using the `message` event.

    :param room: The room that the message was received in.
    :param event: The raw event that triggered the message.
    """
    ...


async def command(ctx: "Context") -> t.Optional[t.Any]:
    """
    This event is dispatched once a command is finished being prepared, and is about to be invoked.

    This event is dispatched *after* the `message` event, but *before* `command_complete` and `command_error`.

    This event features the original context, which can be used to access the message, the command, and the arguments.

    :param ctx: The context of the command.
    """
    ...


async def command_complete(ctx: "Context", result: t.Any) -> t.Optional[t.Any]:
    """
    This event is dispatched after a command has been invoked, and has completed successfully.

    This event features the context, which can be used to access the message, the command, and the arguments.

    :param ctx: The context of the command.
    :param result: The result of the command (the returned value of the callback)
    """
    ...


async def command_error(ctx: "Context", error: "CommandError") -> t.Optional[t.Any]:
    """
    This event is dispatched after a command has been invoked, and has completed with an error.

    This event features the context, which can be used to access the message, the command, and the arguments.

    ??? example "Getting the original error"
        As the error is wrapped in a [niobot.CommandError][], you can access the original error by accessing the
        [`CommandError.original`][niobot.exceptions.CommandError] attribute.

        ```python
        @bot.event("command_error")
        async def on_command_error(ctx, error):
            original_error = error.original
            print("Error:", original_error)
        ```

    It is encouraged that you inform the end user about an error that has occurred, as by default the error is simply
    logged to the console. Don't forget, you've got the whole `Context` instance - use it!

    :param ctx: The context of the command.
    :param error: The error that was raised.
    """
    ...



# OUTDATED AND NOT USED ANYMORE
# Pull from https://github.com/poljar/matrix-nio/pull/451 temporarily.
from nio.responses import DirectRoomsErrorResponse, DirectRoomsResponse

__all__ = (
    "DirectRoomsErrorResponse",
    "DirectRoomsResponse",
)



"""
This example shows you several ways to use command arguments in NioBot.

Basic setup (getting credentials) will not be covered in this example.
"""
from ipaddress import IPv4Address, IPv4Network
from typing import Annotated, Union

import aiohttp
from custom_converters import IPConverter

import niobot

with open("./token.txt") as fd:
    TOKEN = fd.read().strip()


bot = niobot.NioBot(
    "https://matrix.example",
    "@user:matrix.example",
    command_prefix="!"
)


@bot.command("http.get")
async def get_url(
    ctx: niobot.Context,  # Context is never an argument the user can supply.
    url: str  # this is a required argument that will always be a string
):
    """
    This command will fetch the content of the given URL and return it.
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            status = response.status
            content = await response.read()

    return await ctx.respond(f"Status: {status}\nContent: {len(content):,} bytes")


@bot.command("length")
async def length(
    ctx: niobot.Context,
    text: str = None  # this argument is a string, but if not supplied, it will be None.
    # This could also be:
    # text: str | None = None  (python >=3.10)
    # or
    # text: typing.Optional[str] = None  (python 3.9)
):
    """
    This command will return the length of the given text.
    """
    return await ctx.respond(f"Length: {len(text)}")


@bot.command("ip-exists")
async def ip_exists(
    ctx: niobot.Context,
    ip: Annotated[Union[IPv4Address, IPv4Network], IPConverter]
    # `Annotated` allows you to use custom parsers for arguments.
    # The first value of the annotation is the real type, the second is the parser.
    # You *could* just use `ip: IPConverter` but this is better for type-hinting.
):
    """
    This command will return whether the given IP address or network exists.
    """
    await ctx.respond(f"IP: {ip}")
    if isinstance(ip, IPv4Network):
        return await ctx.respond(f"This network has {len(ip.hosts())} possible hosts.")


bot.run(TOKEN)



import re
import textwrap
import typing

from niobot.exceptions import CheckFailure

if typing.TYPE_CHECKING:
    from ..commands import Command
    from ..context import Context


__all__ = ("DefaultHelpCommand",)


class DefaultHelpCommand:
    """
    The default help command for NioBot.

    This is a very basic help command which lists available commands, their arguments, and a short descrption,
    and allows for further information by specifying the command name as an argument.
    """

    def __init__(self):
        pass

    @staticmethod
    def clean_output(
        text: str,
        *,
        escape_user_mentions: bool = True,
        escape_room_mentions: bool = True,
        escape_room_references: bool = False,
        escape_all_periods: bool = False,
        escape_all_at_signs: bool = False,
        escape_method: typing.Optional[typing.Callable[[str], str]] = None,
    ) -> str:
        """
        Escapes given text and sanitises it, ready for outputting to the user.

        This should always be used when echoing any sort of user-provided content, as we all know there will be some
        annoying troll who will just go `@room` for no apparent reason every 30 seconds.

        !!! danger "Do not rely on this!"
            This function is not guaranteed to escape all possible mentions, and should not be relied upon to do so.
            It is only meant to be used as a convenience function for simple commands.

        :param text: The text to sanitise
        :param escape_user_mentions: Escape all @user:homeserver.tld mentions
        :param escape_room_mentions: Escape all @room mentions
        :param escape_room_references: Escape all #room:homeserver.tld references
        :param escape_all_periods: Escape all literal `.` characters (can be used to escape all links)
        :param escape_all_at_signs: Escape all literal `@` characters (can be used to escape all mentions)
        :param escape_method: A custom escape method to use instead of the built-in one
        (which just wraps characters in `\\u200b`)
        :return: The cleaned text
        """
        if escape_method is None:

            def default_escape_method(x: str) -> str:
                return "\u200b".join(x.split())

            escape_method = default_escape_method

        if escape_user_mentions:
            text = re.sub(r"@([A-Za-z0-9\-_=+./]+):([A-Za-z0-9\-_=+./]+)", escape_method("@\\1:\\2"), text)
        if escape_room_mentions:
            text = text.replace("@room", escape_method("@room"))
        if escape_room_references:
            text = re.sub(r"#([A-Za-z0-9\-_=+./]+):([A-Za-z0-9\-_=+./]+)", escape_method("#\\1:\\2"), text)
        if escape_all_periods:
            text = text.replace(".", escape_method("."))
        if escape_all_at_signs:
            text = text.replace("@", escape_method("@"))

        return text

    @staticmethod
    def format_command_name(command: "Command") -> str:
        """Formats the command name with its aliases if applicable"""
        if not command.aliases:
            return command.name
        else:
            return "[{}]".format("|".join([command.name, *command.aliases]))

    def format_command_line(self, prefix: str, command: "Command") -> str:
        """Formats a command line, including name(s) & usage."""
        name = self.format_command_name(command)
        start = f"{prefix}{name}"
        if command.display_usage:
            start += " " + command.display_usage.strip().replace("\n", "")

        return start

    @staticmethod
    def get_short_description(command: "Command") -> str:
        """Generates a short (<100 characters) help description for a command."""
        if not command.description:
            # Get the docstring of the callback
            if command.callback.__doc__:
                # De-indent
                doc = textwrap.dedent(command.callback.__doc__)
            else:
                doc = "No command description."
            description = doc
        else:
            description = command.description

        line = description.splitlines()[0]
        return textwrap.shorten(line, width=100)

    @staticmethod
    def get_long_description(command: "Command") -> str:
        """Gets the full help text for a command."""
        if not command.description:
            # Get the docstring of the callback
            description = command.callback.__doc__ or "No command description."
        else:
            description = command.description

        description = textwrap.dedent(description)

        return "\n".join("> " + x for x in description.splitlines())

    async def get_default_help(self, ctx: "Context", command_name: str = None) -> str:
        """Gets the default help text"""
        lines = []
        prefix = ctx.invoking_prefix or "%prefix%"
        command = None
        if command_name is not None:
            command = ctx.bot.get_command(command_name.casefold())

        if command_name is None:
            added = []
            # Display global help.
            for command in sorted(ctx.client.commands.values(), key=lambda c: c.name):
                if command in added or command.disabled is True:
                    continue  # command is disabled.
                try:
                    await command.can_run(ctx)
                except CheckFailure:
                    continue  # user cannot run command
                display = self.format_command_line(prefix, command)
                description = self.get_short_description(command)
                lines.append("* `{}`: {}".format(display, description))
                added.append(command)
            return "\n".join(lines)
        elif command is None:
            return "No command with the name %r found." % (self.clean_output(command_name))
        else:
            display = self.format_command_line(prefix, command)
            description = self.get_long_description(command)
            lines = ["* {}:".format(display), *description.splitlines()]
            return self.clean_output("\n".join(lines))

    async def respond(self, ctx: "Context", command_name: str = None) -> None:
        """Displays help information about available commands"""
        return await ctx.respond(await self.get_default_help(ctx, command_name))



import typing

import pytest

from niobot.commands import Argument
from niobot.context import Context
from niobot.utils.parsers import *

EXAMPLE_MXC = "mxc://nexy7574.co.uk/6cefe7e807bb846ec44e1a5ec4fdbd5703e932bd46ad500012c16d4d44a79c32"


@pytest.mark.parametrize(
    "parser, input_value, expected_value, ctx, arg",
    [
        (IntegerParser(), "1", 1, None, None),
        (IntegerParser(allow_floats=True), "1.0", 1, None, None),
        (IntegerParser(base=16), "0xFFF", 0xFFF, None, None),
        (FloatParser(), "1.0", 1.0, None, None),
        (FloatParser(), "1", 1.0, None, None),
        (FloatParser(), "1.0e-3", 1.0e-3, None, None),
        (FloatParser(), "1.0e3", 1.0e3, None, None),
        (BooleanParser(), "true", True, None, None),
        (BooleanParser(), "false", False, None, None),
        (BooleanParser(), "1", True, None, None),
        (BooleanParser(), "0", False, None, None),
        (BooleanParser(), "yes", True, None, None),
        (BooleanParser(), "no", False, None, None),
        (BooleanParser(), "on", True, None, None),
        (BooleanParser(), "off", False, None, None),
        (BooleanParser(), "y", True, None, None),
        (BooleanParser(), "n", False, None, None),
        (JSONParser(), '{"a": 1}', {"a": 1}, None, None),
        (JSONParser(), "[1, 2, 3]", [1, 2, 3], None, None),
        (JSONParser(), "1", 1, None, None),
        (JSONParser(), "1.0", 1.0, None, None),
        (JSONParser(), "true", True, None, None),
        (JSONParser(), "false", False, None, None),
        (JSONParser(), "null", None, None, None),
    ],
)
def test_generic_parsers(
    parser: Parser,
    input_value: typing.Any,
    expected_value: typing.Optional[typing.Any],
    ctx: typing.Optional[Context],
    arg: typing.Optional[Argument],
):
    assert parser(ctx, arg, input_value) == expected_value


def test_mxc_parser():
    parser = MXCParser()
    parsed = parser(None, None, EXAMPLE_MXC)
    assert parsed.server == "nexy7574.co.uk"
    assert parsed.media_id == "6cefe7e807bb846ec44e1a5ec4fdbd5703e932bd46ad500012c16d4d44a79c32"


@pytest.mark.parametrize(
    "input_url, expected_value",
    [
        (
            "https://matrix.to/#/!rwJEulKnHLoffvXAof:nexy7574.co.uk/$_dC9O1LkAVxVjZpPowkdg7CbEfUMA3TuWpeLcUzzVa8",
            MatrixToLink("!rwJEulKnHLoffvXAof:nexy7574.co.uk", "$_dC9O1LkAVxVjZpPowkdg7CbEfUMA3TuWpeLcUzzVa8", ""),
        ),
        (
            "https://matrix.to/#/!rwJEulKnHLoffvXAof:nexy7574.co.uk",
            MatrixToLink("!rwJEulKnHLoffvXAof:nexy7574.co.uk", "", ""),
        ),
        ("https://matrix.to/#/@nex:nexy7574.co.uk", MatrixToLink("@nex:nexy7574.co.uk", "", "")),
    ],
)
@pytest.mark.asyncio
async def test_matrix_to_parser(input_url: str, expected_value: typing.Any):
    parser = MatrixToParser(stateless=True)
    parsed = await parser(None, None, input_url)
    for n in range(len(parsed)):
        assert parsed[n] == expected_value[n]



import asyncio
import importlib.metadata
import logging
import pathlib
import platform
import re
import shutil
import sys
import typing

import packaging.specifiers
import packaging.version

import niobot

try:
    import click
    import httpx
    import psutil
except ImportError:
    print("Missing CLI dependencies. Did you install CLI extras?", file=sys.stderr)
    sys.exit(1)

logger = logging.getLogger("niobot.cli")


def versions(package_name: str) -> typing.List[packaging.version.Version]:
    response = httpx.get(
        "https://pypi.org/simple/%s/" % package_name,
        headers={
            "User-Agent": "nio-bot (CLI, https://github.com/nexy7574/niobot)",
            "Accept": "application/vnd.pypi.simple.v1+json",
        },
    )
    assert response.status_code == 200
    assert response.headers["Content-Type"] == "application/vnd.pypi.simple.v1+json"
    return list(map(packaging.version.parse, response.json()["versions"]))


@typing.overload
def get_installed_version(package_name: str, default_to_0_0_0: bool = True, *, parse: typing.Literal[False]) -> str: ...


@typing.overload
def get_installed_version(
    package_name: str, default_to_0_0_0: bool = True, *, parse: typing.Literal[True]
) -> packaging.version.Version: ...


def get_installed_version(
    package_name: str, default_to_0_0_0: bool = True, *, parse: bool = False
) -> typing.Union[str, packaging.version.Version]:
    """
    Fetches a package version via importlib.metadata.

    :param package_name: The name of the package to fetch.
    :param default_to_0_0_0: Whether to default to "0.0.0" if the package is not installed or not found.
    :return: The package version identifier or packaging version
    """
    try:
        version = importlib.metadata.version(package_name)
    except importlib.metadata.PackageNotFoundError:
        if default_to_0_0_0:
            version = "0.0.0"
        raise

    if parse:
        version = packaging.version.parse(version)
    return version


@click.group()
@click.option(
    "--log-level",
    "-L",
    type=click.Choice(["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"], case_sensitive=False),
    default="WARNING",
)
@click.pass_context
def cli_root(ctx, log_level: str):
    """CLI utilities for nio-bot."""
    logging.basicConfig(
        level=getattr(logging, log_level),
    )


@cli_root.command()
@click.option("--no-colour", "--no-color", "-C", is_flag=True, default=False)
def version(no_colour: bool):
    """Shows version information."""
    logger.info("Gathering version info...")
    import platform

    from nio.crypto import ENCRYPTION_ENABLED

    try:
        nio_version = get_installed_version("matrix-nio", parse=True)
    except (importlib.metadata.PackageNotFoundError, packaging.version.InvalidVersion) as e:
        logger.critical("Failed to resolve matrix-nio version information", exc_info=e)
        nio_version = packaging.version.Version("0.0.0")

    try:
        niobot_version = get_installed_version("nio-bot", parse=True)

        # Check for updates
        niobot_versions = versions("nio-bot")
        allow_pre = niobot_version.is_prerelease
        newest = tuple(filter(lambda v: v.is_prerelease in (False, allow_pre), niobot_versions))[-1]
        if newest > niobot_version:
            msg = "There is an update to nio-bot available: %s" % newest
            logger.warning(msg)
            click.secho(msg, fg="yellow")
    except (importlib.metadata.PackageNotFoundError, packaging.version.InvalidVersion) as e:
        logger.critical("Failed to resolve nio-bot version information", exc_info=e)
        niobot_version = packaging.version.Version("0.0.0")

    is_release_version = not any(
        (niobot_version.is_devrelease, niobot_version.is_prerelease, niobot_version.is_postrelease)
    )
    niobot_version_pretty = niobot_version.public
    if not is_release_version:
        build = niobot_version.base_version
        dev_build = niobot_version.dev or "N/A"
        post = niobot_version.post or "N/A"
        pre = "".join(map(str, niobot_version.pre or ("N/A",)))
        commit = niobot_version.local or "N/A"

        niobot_version_pretty += f" (v{build}, build {dev_build}, pre {pre}, post {post}, commit {commit})"

    _os = platform.platform()
    if sys.version_info > (3, 9):
        if hasattr(platform, "freedesktop_os_release"):
            try:
                _os_info = platform.freedesktop_os_release()
            except OSError:
                pass
            else:
                _os += " ({0}/{1} - {2})".format(
                    _os_info.get("NAME", "Unknown"),
                    _os_info.get("VERSION", "Unknown"),
                    _os_info.get("PRETTY_NAME", "Unknown"),
                )

    space = {"total": 0}
    for mount in psutil.disk_partitions():
        try:
            usage = shutil.disk_usage(mount.mountpoint)
            free_percent = (usage.free / usage.total) * 100
            space["total"] += usage.free
            space[mount.mountpoint] = round(free_percent, 2)
        except PermissionError:
            pass
    space_total = space.pop("total")

    supported_python_versions_raw = importlib.metadata.metadata("nio-bot")["Requires-Python"]
    supported_python_versions = packaging.specifiers.SpecifierSet(supported_python_versions_raw)
    if not supported_python_versions.contains(platform.python_version()):
        python_version_supported = False
    else:
        python_version_supported = True

    nio_versions_raw = ""
    for version in importlib.metadata.requires("nio-bot"):
        if version.startswith("matrix-nio"):
            nio_versions_raw = "~" + version.split("~")[1]
            break
    nio_versions = packaging.specifiers.SpecifierSet(nio_versions_raw)
    if not nio_versions.contains(nio_version):
        nio_version_supported = False
    else:
        nio_version_supported = True

    lines = [
        ["NioBot version", niobot_version_pretty, lambda x: True],
        ["matrix-nio version", nio_version, lambda x: nio_version_supported],
        ["Python version", platform.python_version(), lambda x: python_version_supported],
        ["Python implementation", platform.python_implementation(), lambda x: x == "CPython"],
        ["Operating System", _os, lambda val: val.startswith(("Windows", "Linux"))],
        ["Architecture", platform.machine(), lambda x: x == "x86_64"],
        ["OLM Installed", "Yes" if ENCRYPTION_ENABLED else "No", lambda x: x != "No"],
        ["Free Disk Space", " ".join(f"{k} ({v}%)" for k, v in space.items()), lambda x: space_total > 1024**3],
    ]

    _docker_path = pathlib.Path("/.dockerenv")
    if _docker_path.is_file():
        lines.append(["Running in Docker", "Probably (/.dockerenv exists)", lambda _: True])

    click.echo()
    for line in lines:
        if no_colour:
            click.echo("%s: %s" % tuple(line)[:2])
        else:
            click.echo(
                click.style(line[0], fg="cyan") + ": " + click.style(line[1], fg="green" if line[2](line[1]) else "red")
            )
    click.echo()


@cli_root.command(name="test-homeserver")
@click.argument("homeserver")
def test_homeserver(homeserver: str):
    """Walks through resolving and testing a given homeserver."""
    import urllib.parse

    import httpx

    if homeserver.startswith("@"):
        _, homeserver = homeserver.split(":", 1)

    logger.debug("Parsing given input %r as a URL...", homeserver)
    parsed = urllib.parse.urlparse(homeserver)
    if not parsed.scheme:
        logger.debug("No scheme found, assuming HTTPS.")
        parsed = urllib.parse.urlparse(f"https://{homeserver}")

    if not parsed.netloc:
        logger.critical("No netloc found, cannot continue.")
        logger.critical("URI parsed to: %r", parsed)
        return

    logger.debug("Attempting to resolve %r...", parsed.netloc)
    logger.info("Trying well-known of %r...", parsed.netloc)
    base_url = None
    try:
        response = httpx.get(f"https://{parsed.netloc}/.well-known/matrix/client", timeout=30, follow_redirects=True)
        response.raise_for_status()
    except httpx.HTTPError as e:
        logger.critical("Failed to get well-known: %r", e)
        return
    else:
        logger.debug("Got response: %r", response)
        cors = response.headers.get("Access-Control-Allow-Origin", None)
        if not cors:
            logger.warning("No CORS header found. This homeserver will be unusable in web-based clients!")
        elif cors != "*":
            logger.warning(
                "CORS header found, but not set to wildcard. "
                "This homeserver will be unusable in most web-based clients!"
            )

        if response.status_code != 404:
            if response.status_code != 200 or len(response.content or b"") == 0:
                logger.critical("Well-known returned non-404, but no content. Failing.")
                return
            else:
                data = response.json()
                logger.debug("Well-known data: %r", data)
                if "m.homeserver" not in data:
                    logger.critical("Well-known data does not contain m.homeserver. Failing.")
                    return
                base_url = urllib.parse.urlparse(data["m.homeserver"]["base_url"])
                logger.debug("Parsed base URL: %r", base_url)
                if not base_url.scheme or not base_url.netloc:
                    logger.critical("Base URL does not contain a scheme. Invalid URL? Failing.")
                    return

    if not base_url:
        logger.info("No well-known found. Assuming %r as homeserver.", parsed.netloc)
        base_url = urllib.parse.urlparse(f"https://{parsed.netloc}")

    base_url = base_url.geturl()
    logger.info("Using %r as homeserver.", base_url)
    logger.info("Validating homeserver...")
    try:
        response = httpx.get(f"{base_url}/_matrix/client/versions", timeout=30)
    except httpx.HTTPError as e:
        logger.critical("Failed to get versions: %r", e)
        return
    else:
        data = response.json()
        logger.debug("Got response: %r", data)
        if "versions" not in data:
            logger.critical("Versions response does not contain versions. Failing.")
            return
        if "unstable_features" not in data:
            logger.warning("Versions response does not contain unstable_features. This may be bad.")

        logger.info("Homeserver validated.")
        click.secho("Homeserver validated.", fg="green")
        click.secho("%s -> %s" % (click.style(parsed.netloc, dim=True), click.style(base_url, fg="green")))


@cli_root.command(name="get-access-token")
@click.option("--username", "-U", default=None, help="The username part (without @)")
@click.option("--password", "-P", default=None, help="The password to use (will be prompted if not given)")
@click.option("--homeserver", "-H", default=None, help="The homeserver to use (will be prompted if not given)")
@click.option(
    "--device-id",
    "-D",
    "--session",
    "--session-id",
    default=None,
    help="The device ID to use (will be prompted if not given)",
)
@click.option(
    "--output", "-o", "-O", default="-", help="The file to output the results to.", type=click.Path(allow_dash=True)
)
def get_access_token(output: str, username: str, password: str, homeserver: str, device_id: str):
    """Fetches your access token from your homeserver."""
    if not username:
        username = ""

    while not re.match(r"@[\w\-.]{1,235}:[0-9A-Za-z\-_.]{3,253}", username):
        username = click.prompt("User ID (@username:homeserver.tld)")
        _, homeserver = username.split(":", 1)

    if not homeserver:
        _, homeserver = username.split(":", 1)

    if not password:
        password = click.prompt("Password (will not echo)", hide_input=True)

    if not homeserver:
        homeserver = click.prompt("Homeserver URL")

    if not device_id:
        node = platform.node()
        device_id = click.prompt(
            "Device ID (a memorable display name for this login, such as 'bot-production')", default=node
        )

    click.secho("Resolving homeserver... ", fg="cyan", nl=False)
    try:
        homeserver = asyncio.run(niobot.resolve_homeserver(homeserver))
    except ConnectionError:
        click.secho("Failed!", fg="red")
    else:
        click.secho("OK", fg="green")

    click.secho("Getting access token... ", fg="cyan", nl=False)
    status_code = None
    try:
        response = httpx.post(
            f"{homeserver}/_matrix/client/r0/login",
            json={
                "type": "m.login.password",
                "identifier": {"type": "m.id.user", "user": username},
                "password": password,
                "device_id": device_id,
                "initial_device_display_name": device_id,
            },
        )
        status_code = response.status_code
        if status_code == 429:
            click.secho("Failed!", fg="red", nl=False)
            click.secho(" (Rate limited for {:.0f} seconds)".format(response.json()["retry_after_ms"] / 1000), bg="red")
            return
        response.raise_for_status()
    except httpx.HTTPError as e:
        click.secho("Failed!", fg="red", nl=False)
        click.secho(f" ({status_code or str(e)})", bg="red")
    else:
        click.secho("OK", fg="green")
        if output != "-":
            click.secho(f'Access token: {response.json()["access_token"]}', fg="green")
            with click.open_file(output, "w+") as f:
                f.write(response.json()["access_token"])
        else:
            click.echo(response.json()["access_token"])


if __name__ == "__main__":
    cli_root()



from __future__ import annotations

import logging
import time
import typing

import nio

from .utils.lib import deprecated
from .utils.string_view import ArgumentView

if typing.TYPE_CHECKING:
    from .attachment import BaseAttachment
    from .client import NioBot
    from .commands import Command


__all__ = ("Context", "ContextualResponse")

logger = logging.getLogger(__name__)


class ContextualResponse:
    """Context class for managing replies.

    Usage of this function is not required, however it is a useful utility."""

    def __init__(self, ctx: "Context", response: nio.RoomSendResponse):
        self.ctx = ctx
        self._response = response

    def __repr__(self):
        return "<ContextualResponse ctx={0.ctx!r} response={0.response!r}>".format(self)

    @property
    @deprecated("original_event")
    def message(self) -> typing.Optional[nio.RoomMessage]:
        result = self.ctx.client.get_cached_message(self._response.event_id)
        if result:
            return result[1]
        else:
            logger.warning("Original response for context %r was not found in cache, unable to modify.", self.ctx)

    async def original_event(self) -> typing.Optional[nio.RoomMessage]:
        """Fetches the current event for this response"""
        result = self.ctx.client.get_cached_message(self._response.event_id)
        if result:
            return result[1]
        event = await self.ctx.client.room_get_event(self.ctx.room.room_id, self._response.event_id)
        if isinstance(event, nio.RoomGetEventResponse):
            return nio.RoomMessage.parse_event(event.event.source)

    async def reply(
        self,
        content: typing.Optional[str] = None,
        file: typing.Optional[BaseAttachment] = None,
        message_type: typing.Optional[str] = None,
        *,
        content_type: typing.Literal["plain", "markdown", "html", "html.raw"] = "markdown",
        override: typing.Optional[dict] = None,
    ) -> "ContextualResponse":
        """
        Replies to the current response.

        This does NOT reply to the original invoking message.

        See [niobot.NioBot.send_message][] for more information.
        """
        return ContextualResponse(
            self.ctx,
            await self.ctx.client.send_message(
                self.ctx.room,
                content,
                file,
                message_type=message_type,
                content_type=content_type,
                reply_to=self._response.event_id,
                override=override,
            ),
        )

    async def edit(
        self,
        content: str,
        *,
        message_type: typing.Optional[str] = None,
        content_type: typing.Literal["plain", "markdown", "html", "html.raw"] = "markdown",
        override: typing.Optional[dict] = None,
    ) -> "ContextualResponse":
        """
        Edits the current response.

        See [niobot.NioBot.edit_message][] for more information.
        """
        await self.ctx.client.edit_message(
            self.ctx.room,
            self._response.event_id,
            content,
            message_type=message_type,
            content_type=content_type,
            override=override,
        )
        return self

    async def delete(self, reason: typing.Optional[str] = None) -> None:
        """
        Redacts the current response.

        :param reason: An optional reason for the redaction
        :return: None, as there will be no more response.
        """
        await self.ctx.client.delete_message(self.ctx.room, self._response.event_id, reason=reason)


class Context:
    """Event-based context for a command callback"""

    def __init__(
        self,
        _client: "NioBot",
        room: nio.MatrixRoom,
        event: nio.RoomMessageText,
        command: "Command",
        *,
        invoking_prefix: typing.Optional[str] = None,
        invoking_string: typing.Optional[str] = None,
    ):
        self._init_ts = time.time()
        self._client = _client
        self._room = room
        self._event = event
        self._command = command
        self.invoking_prefix = invoking_prefix
        self._invoking_string = invoking_string
        to_parse = event.body

        if invoking_string:
            try:
                to_parse = event.body[len(invoking_string) :]
            except IndexError:
                to_parse = ""
        self._args = ArgumentView(to_parse)
        self._args.parse_arguments()
        self._original_response = None

        # property aliases
        self.bot = self.client
        self.msg = self.event = self.message
        self.arguments = self.args

    def __repr__(self):
        return "<Context room={0.room!r} event={0.event!r} command={0.command!r}>".format(self)

    def __eq__(self, other):
        if not isinstance(other, Context):
            return False
        return self.room == other.room and self.event == other.event and self.command == other.command

    @property
    def room(self) -> nio.MatrixRoom:
        """The room that the event was dispatched in"""
        return self._room

    @property
    def client(self) -> "NioBot":
        """The current instance of the client"""
        return self._client

    @property
    def command(self) -> "Command":
        """The current command being invoked"""
        return self._command

    @property
    def args(self) -> list[str]:
        """Each argument given to this command"""
        return self._args.arguments

    @property
    def message(self) -> nio.RoomMessageText:
        """The current message"""
        return self._event

    @property
    def original_response(self) -> typing.Optional[nio.RoomSendResponse]:
        """The result of Context.reply(), if it exists."""
        return self._original_response

    @property
    def latency(self) -> float:
        """Returns the current event's latency in milliseconds."""
        return self.client.latency(self.event, received_at=self._init_ts)

    async def respond(
        self,
        content: typing.Optional[str] = None,
        file: typing.Optional[BaseAttachment] = None,
        reply_to: typing.Optional[typing.Union[nio.RoomMessageText, str]] = None,
        message_type: typing.Optional[str] = None,
        *,
        content_type: typing.Literal["plain", "markdown", "html", "html.raw"] = "markdown",
        override: typing.Optional[dict] = None,
    ) -> ContextualResponse:
        """
        Responds to the current event.

        See [niobot.NioBot.send_message][] for more information.
        """
        result = await self.client.send_message(
            self.room,
            content,
            file,
            reply_to=reply_to or self.event,
            message_type=message_type,
            content_type=content_type,
            override=override,
        )
        return ContextualResponse(self, result)



from ipaddress import IPv4Address, IPv4Network

from niobot import CommandArgumentsError, StatelessParser


class IPConverter(StatelessParser):
    """
    Converts an input into either an IP(v4) address or IP network.
    """
    def __call__(self, value: str) -> IPv4Address | IPv4Network:
        try:
            return IPv4Address(value)
        except ValueError:
            try:
                return IPv4Network(value)
            except ValueError as e:
                raise CommandArgumentsError("Invalid IP address or network.", exception=e)



"""
This utility modules contains a handful of simple off-the-shelf parser for some basic python types.
"""

import abc
import re
import typing
import urllib.parse as urllib
from collections import namedtuple
from typing import Union as U

import nio

from ..exceptions import CommandParserError
from .lib import deprecated

if typing.TYPE_CHECKING:
    from ..commands import Argument
    from ..context import Context

try:
    import orjson as json
except ImportError:
    # Orjson is not installed.
    import json


__all__ = (
    "Parser",
    "StatelessParser",
    "BooleanParser",
    "FloatParser",
    "IntegerParser",
    "JSONParser",
    "EventParser",
    "RoomParser",
    "BUILTIN_MAPPING",
    "MATRIX_TO_REGEX",
    "MatrixToLink",
    "MatrixMXCUrl",
    "MatrixToParser",
    "MXCParser",
    # And the deprecated aliases
    "boolean_parser",
    "float_parser",
    "integer_parser",
    "json_parser",
    "room_parser",
    "event_parser",
    "matrix_to_parser",
    "mxc_parser",
)
MATRIX_TO_REGEX = re.compile(
    r"(http(s)?://)?matrix\.to/#/(?P<room_id>[^/]+)(/(?P<event_id>[^/?&#]+))?(?P<qs>([&?](via=[^&]+))*)?",
)
MatrixToLink = namedtuple("MatrixToLink", ("room", "event", "query"), defaults=("", "", ""))
MatrixMXCUrl = namedtuple("MatrixMXCUrl", ("server", "media_id"), defaults=(None, None))


class Parser(abc.ABC):
    """
    A base model class for parsers.

    This ABC defines one uniform method, which is `__call__`, which takes a [Context][niobot.context.Context] instance,
    [Argument][niobot.commands] instance, and the user-provided string value.

    This parser is designed to be instantiated, and then called with the above arguments.
    If you want to make a simple parser that does not take additional configuration, it is recommended to use
    [StatelessParser][niobot.utils.parsers.StatelessParser] instead.
    """

    @abc.abstractmethod
    def __call__(self, ctx: "Context", arg: "Argument", value: str) -> typing.Optional[typing.Any]: ...


class StatelessParser(Parser, abc.ABC):
    r"""
    A parser base that will not be instantiated, but rather called directly.

    This is useful for parsers that do not take any configuration
    (such as the simple [BooleanParser][niobot.utils.parsers]), where a
    simple one-off call is enough.

    Traditionally, you'd call a [Parser][niobot.utils.parsers] like this:

    ```python
    parser = Parser(my_argument=True)
    result = parser(ctx, arg, value)
    # or, in one line
    result = Parser(my_argument=True)(ctx, arg, value)
    ```

    However, for some simple parsers, there's no need to instantiate them. Instead, you can call them directly.
    The `StatelessParser` ABC adds the `parse` classmethod, meaning you can simply do the following:

    ```python
    result = Parser.parse(ctx, arg, value)
    ```
    Which is just a shortand for the above one-liner.
    This offers little to no performance benefit, however can make code look cleaner.

    As this ABC subclasses the regular [Parser][niobot.utils.parsers],
    you can still use the traditional instantiation+call method.
    """

    @classmethod
    def parse(cls, ctx: "Context", arg: "Argument", value: str) -> typing.Optional[typing.Any]:
        r"""Parses the given value using this parser without needing to call \_\_init\_\_() first.

        :param ctx: The context instance
        :param arg: The argument instance
        :param value: The value to parse
        :return: The parsed value
        :rtype: typing.Optional[typing.Any]"""
        return cls()(ctx, arg, value)


class BooleanParser(StatelessParser):
    """
    Converts a given string into a boolean. Value is casefolded before being parsed.

    The following resolves to true:
    * 1, y, yes, true, on

    The following resolves to false:
    * 0, n, no, false, off

    The following will raise a command argument error: anything else

    :return: A parsed boolean
    :rtype: bool
    """

    def __call__(self, ctx: "Context", arg: "Argument", value: str) -> bool:
        value = value.casefold()
        if value in {"1", "y", "yes", "true", "on"}:
            return True
        if value in {"0", "n", "no", "false", "off"}:
            return False
        raise CommandParserError(f"Invalid boolean value: {value}. Should be a sensible value, such as 1, yes, false.")


class FloatParser(StatelessParser):
    """
    Converts a given string into a floating point number.

    :return: A parsed floating point number
    :rtype: float
    :raises CommandParserError: if the value is not a valid number.
    """

    def __call__(self, ctx: "Context", arg: "Argument", value: str) -> float:
        try:
            return float(value)
        except ValueError:
            raise CommandParserError(f"Invalid float value: {value}. Should be a number.")


class IntegerParser(Parser):
    """
    Parses an integer, or optionally a real number.

    :param allow_floats: Whether to simply defer non-explicit-integer values to the float parser.
        This results in the return type being [float][]
    :param base: The base to parse the integer in. Defaults to 10 (denary). 2 is Binary, and 16 is Hexadecimal.
    :return: A parsed integer or float, depending on input & allow_floats
    :rtype: Union[int, float]
    :raises CommandParserError: if the value is not a valid number.
    """

    def __init__(self, allow_floats: bool = False, base: int = 10):
        self.allow_floats = allow_floats
        self.base = base

    def __call__(self, ctx: "Context", arg: "Argument", value: str) -> U[int, float]:
        try:
            return int(value, self.base)
        except ValueError:
            if self.allow_floats:
                try:
                    return float(value)
                except ValueError as e:
                    raise CommandParserError(f"Invalid integer value: {value}. Should be a number.", exception=e)
            raise CommandParserError(f"Invalid integer value: {value}. Should be a number.")


class JSONParser(StatelessParser):
    """
    Converts a given string into a JSON object.

    !!! note "Performance boost"
        If you want this to be fast, you should install orjson. It is a drop-in replacement for the standard library.
        While the parser will still work without it, it may be slower, especially for larger payloads.

    :return: The parsed JSON object
    :rtype: Union[dict, list, str, int, float, None, bool]
    :raises CommandParserError: if the value is not a valid JSON object.
    """

    def __call__(self, ctx: "Context", arg: "Argument", value: str) -> typing.Any:
        try:
            return json.loads(value)
        except json.JSONDecodeError as e:
            raise CommandParserError(f"Invalid JSON value: {value}. Should be a valid JSON object.", exception=e)


class RoomParser(StatelessParser):
    """
    Parses a room ID, alias, or matrix.to link into a MatrixRoom object.

    ??? warning "This parser is async"
        This parser is async, and should be awaited when used manually.

    :return: The parsed room instance
    :rtype: nio.MatrixRoom
    """

    @staticmethod
    async def internal(ctx: "Context", value: str) -> nio.MatrixRoom:
        if value.startswith("!"):
            # Room ID
            room = ctx.client.rooms.get(value)
        elif value.startswith("#"):
            # Room alias
            # Attempt to find the room in the cache, before requesting it from the server
            for room in ctx.client.rooms.values():
                room: nio.MatrixRoom
                if value == room.canonical_alias:
                    break
            else:
                room: U[nio.RoomResolveAliasResponse, nio.RoomResolveAliasError] = await ctx.client.room_resolve_alias(
                    value
                )
                if isinstance(room, nio.RoomResolveAliasError):
                    raise CommandParserError(f"Invalid room alias: {value}.", response=room)
                room = ctx.client.rooms.get(room.room_id)
        elif m := MATRIX_TO_REGEX.match(value):
            # matrix.to link
            groups = m.groupdict()
            if "room_id" not in groups:
                raise CommandParserError(f"Invalid matrix.to link: {value} (no room).")
            room = ctx.client.rooms.get(groups["room_id"])
        else:
            raise CommandParserError(f"Invalid room ID, alias, or matrix.to link: {value!r}.")

        if room is None:
            raise CommandParserError("No room with that ID, alias, or matrix.to link found.")
        return room

    def __call__(self, ctx, arg, value: str) -> typing.Coroutine[typing.Any, typing.Any, nio.MatrixRoom]:
        return self.internal(ctx, value)


class EventParser(Parser):
    """
    Parses an event reference from either its ID, or matrix.to link.

    :param event_type: The event type to expect (such as m.room.message). If None, any event type is allowed.
    :return: The actual internal (async) parser.
    :rtype: typing.Coroutine
    """

    def __init__(self, event_type: typing.Optional[str] = None):
        self.event_type = event_type

    async def internal(self, ctx: "Context", value: str) -> nio.Event:
        room_id = ctx.room.room_id
        if m := MATRIX_TO_REGEX.match(value):
            # matrix.to link
            groups = m.groupdict()
            if "room_id" not in groups:
                raise CommandParserError(f"Invalid matrix.to link: {value} (no room).")
            if "event_id" not in groups:
                raise CommandParserError(f"Invalid matrix.to link: {value} (expected an event).")
            value = groups["event_id"]
            room_id = groups["room_id"]
            value = urllib.unquote(value)
            room_id = urllib.unquote(room_id)

        if value.startswith("$"):
            # from raw ID
            event: U[nio.RoomGetEventResponse, nio.RoomGetEventError] = await ctx.client.room_get_event(room_id, value)
            if not isinstance(event, nio.RoomGetEventResponse):
                raise CommandParserError(f"Invalid event ID: {value}.", response=event)
            event: nio.Event = event.event
            if self.event_type is not None and self.event_type != event.source.get("type"):
                raise CommandParserError(
                    f"Invalid event ID: {value} (expected {self.event_type}, got {event.source.get('type')})."
                )
            return event
        else:
            raise CommandParserError(f"Invalid event ID or matrix.to link: {value!r}.")

    def __call__(self, ctx, arg, value) -> typing.Coroutine[typing.Any, typing.Any, nio.Event]:
        return self.internal(ctx, value)


class MatrixDotToParser(Parser):
    """
    Converts a matrix.to link into a MatrixRoomLink namedtuple, which consists of the room, event, and any query
    passed to the URL.

    :param domain: The domain to check for. Defaults to matrix.to, consistent with average client behaviour.
    :param require_room: Whether to require the room part of this url to be present
    :param require_event: Whether to require the event part of this url to be present
    :param allow_user_as_room: Whether to allow user links as room links
    :param stateless: If true, the link will only be parsed, not resolved. This means rooms will stay as their IDs, etc.
    :return: The actual internal (async) parser.
    :rtype: typing.Coroutine
    """

    def __init__(
        self,
        domain: str = "matrix.to",
        require_room: bool = True,
        require_event: bool = False,
        allow_user_as_room: bool = True,
        *,
        stateless: bool = False,
    ):
        self.domain = domain
        self.require_room = require_room
        self.require_event = require_event
        self.allow_user_as_room = allow_user_as_room
        self.stateless = stateless

    async def internal(self, ctx: "Context", value: str) -> MatrixToLink:
        if not (m := MATRIX_TO_REGEX.match(value)):
            raise CommandParserError(f"Invalid matrix.to link: {value!r}.")

        # matrix.to link
        groups = m.groupdict()
        event_id = groups.get("event_id", "")
        room_id = groups.get("room_id", "")
        event_id = urllib.unquote(event_id or "")
        room_id = urllib.unquote(room_id or "")

        if self.require_room and not room_id:
            raise CommandParserError(f"Invalid matrix.to link: {value} (no room).")
        if self.require_event and not event_id:
            raise CommandParserError(f"Invalid matrix.to link: {value} (no event).")

        if room_id.startswith("@") and not self.allow_user_as_room:
            raise CommandParserError(f"Invalid matrix.to link: {value} (expected room, got user).")

        if self.stateless:
            room = room_id
        else:
            if room_id.startswith("@"):
                dm_rooms = await ctx.client.get_dm_rooms(room_id)
                if not dm_rooms:
                    new_dm_room = await ctx.client.create_dm_room(room_id)
                    dm_rooms.append(new_dm_room.room_id)
                room_id = dm_rooms[0]
            room = ctx.client.rooms.get(room_id)
            if room is None:
                raise CommandParserError("No room with that ID, alias, or matrix.to link found.")

        if event_id and not self.stateless:
            event: U[nio.RoomGetEventResponse, nio.RoomGetEventError] = await ctx.client.room_get_event(
                room_id, event_id
            )
            if not isinstance(event, nio.RoomGetEventResponse):
                raise CommandParserError(f"Invalid event ID: {event_id}.", response=event)
            event: nio.Event = event.event
        elif self.stateless:
            event: str = event_id
        else:
            event: None = None
        return MatrixToLink(room, event, groups.get("qs"))

    def __call__(
        self, ctx: "Context", arg: "Argument", value: str
    ) -> typing.Coroutine[typing.Any, typing.Any, MatrixToLink]:
        return self.internal(ctx, value)


MatrixToParser = MatrixDotToParser


class MXCParser(StatelessParser):
    """
    Parses an MXC URL into a MatrixMXCUrl namedtuple, which consists of the server and media ID.

    :return: The parsed MXC URL
    :rtype: MatrixMXCUrl (namedtuple)
    """

    def __call__(self, ctx, arg, value: str) -> MatrixMXCUrl:
        if not value.startswith("mxc://"):
            raise CommandParserError(f"Invalid MXC URL: {value!r}.")
        parsed = urllib.urlparse(value, "mxc://", allow_fragments=False)
        if not parsed.path:
            raise CommandParserError(f"Invalid MXC URL: {value!r} (missing media ID).")
        if parsed.path == "/":
            raise CommandParserError(f"Invalid MXC URL: {value!r} (no media ID).")
        if not parsed.netloc:
            raise CommandParserError(f"Invalid MXC URL: {value!r} (no server).")

        return MatrixMXCUrl(parsed.netloc, parsed.path[1:])


class MatrixUserParser(StatelessParser):
    """
    Parses a string into a MatrixUser instance from matrix-nio.
    """

    def __call__(self, ctx, arg, value):
        if not value.startswith("@"):
            raise CommandParserError(f"Invalid matrix user ID: {value!r}.")

        # Check the local room for the user
        for member in ctx.room.users.values():
            if member.user_id == value:
                return member
        raise CommandParserError(f"Invalid matrix user ID: {value!r} (not in room).")


BUILTIN_MAPPING = {
    bool: BooleanParser(),
    float: FloatParser(),
    int: IntegerParser(),
    list: JSONParser(),
    dict: JSONParser(),
    nio.RoomMessage: EventParser("m.room.message"),
    nio.Event: EventParser(),
    nio.MatrixRoom: RoomParser(),
    nio.MatrixUser: MatrixUserParser(),
    MatrixToLink: MatrixDotToParser(),
}
# Now add the aliases for backward compatability with <1.1.0
# These will be removed in 1.2.0.


@deprecated("niobot.utils.parsers.BooleanParser")
def boolean_parser(*args, **kwargs):
    """Deprecated boolean parser. Please use niobot.utils.parsers.BooleanParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return BooleanParser.parse(*args, **kwargs)


@deprecated("niobot.utils.parsers.FloatParser")
def float_parser(*args, **kwargs):
    """Deprecated float parser. Please use niobot.utils.parsers.FloatParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return FloatParser.parse(*args, **kwargs)


@deprecated("niobot.utils.parsers.IntegerParser")
def integer_parser(allow_floats: bool = False, base: int = 10):
    """Deprecated integer parser. Please use niobot.utils.parsers.IntegerParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return IntegerParser(allow_floats, base)


@deprecated("niobot.utils.parsers.JSONParser")
def json_parser(*args, **kwargs):
    """Deprecated integer parser. Please use niobot.utils.parsers.JSONParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return JSONParser.parse(*args, **kwargs)


@deprecated("niobot.utils.parsers.RoomParser")
def room_parser(*args, **kwargs):
    """Deprecated room parser. Please use niobot.utils.parsers.RoomParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return RoomParser.parse(*args, **kwargs)


@deprecated("niobot.utils.parsers.EventParser")
def event_parser(event_type: typing.Optional[str] = None):
    """Deprecated event parser. Please use niobot.utils.parsers.EventParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return EventParser(event_type)


@deprecated("niobot.utils.parsers.MatrixDotToParser")
def matrix_to_parser(require_room: bool = True, require_event: bool = False, allow_user_as_room: bool = True):
    """Deprecated matrix.to parser. Please use niobot.utils.parsers.MatrixDotToParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return MatrixDotToParser("matrix.to", require_room, require_event, allow_user_as_room)


@deprecated("niobot.utils.parsers.MXCParser")
def mxc_parser(*args, **kwargs):
    """Deprecated MXC parser. Please use niobot.utils.parsers.MXCParser instead.

    !!! danger "Deprecated function"
        This function is deprecated and will be removed in 1.2.0.
    """
    return MXCParser.parse(*args, **kwargs)



import asyncio
import getpass
import importlib
import inspect
import json
import logging
import os
import pathlib
import re
import sys
import time
import typing
import warnings
from collections import deque
from typing import Optional, Union as U

import marko
import nio
from nio.crypto import ENCRYPTION_ENABLED

from .attachment import BaseAttachment
from .commands import Command, Module
from .exceptions import (
    CheckFailure,
    CommandArgumentsError,
    CommandDisabledError,
    CommandError,
    GenericMatrixError,
    LoginException,
    MessageException,
    NioBotException,
)
from .utils import Mentions, Typing, deprecated, force_await, run_blocking
from .utils.help_command import DefaultHelpCommand

try:
    from bs4 import BeautifulSoup
except ImportError:
    BeautifulSoup = None

if typing.TYPE_CHECKING:
    from .context import Context

__all__ = ("NioBot",)

T = typing.TypeVar("T")


class NioBot(nio.AsyncClient):
    """
    The main client for NioBot.

    :param homeserver: The homeserver to connect to. e.g. https://matrix-client.matrix.org
    :param user_id: The user ID to log in as. e.g. @user:matrix.org
    :param device_id: The device ID to log in as. e.g. nio-bot
    :param store_path: The path to the store file. Defaults to ./store. Must be a directory.
    :param command_prefix: The prefix to use for commands. e.g. `!`. Can be a string, a list of strings,
     or a regex pattern.
    :param case_insensitive: Whether to ignore case when checking for commands. If True, this casefold()s
     incoming messages for parsing.
    :param global_message_type: The message type to default to. Defaults to m.notice
    :param ignore_old_events: Whether to simply discard events before the bot's login.
    :param auto_join_rooms: Whether to automatically join rooms the bot is invited to.
    :param auto_read_messages: Whether to automatically update read recipts
    :param owner_id: The user ID of the bot owner. If set, only this user can run owner-only commands, etc.
    :param max_message_cache: The maximum number of messages to cache. Defaults to 1000.
    :param ignore_self: Whether to ignore messages sent by the bot itself. Defaults to False. Useful for self-bots.
    :param import_keys: A key export file and password tuple. These keys will be imported at startup.
    """

    # Long typing definitions out here instead of in __init__ to just keep it cleaner.
    _events: typing.Dict[typing.Union[str, nio.Event], typing.List[typing.Callable[..., typing.Any]]]
    """Internal events register."""
    _commands: typing.Dict[str, Command]
    """Internal command register."""
    _modules: typing.Dict[typing.Type, Module]
    """Internal module register."""

    def __init__(
        self,
        homeserver: str,
        user_id: str,
        device_id: str = "nio-bot",
        store_path: typing.Optional[str] = None,
        *,
        command_prefix: typing.Union[str, re.Pattern, typing.Iterable[str]],
        case_insensitive: bool = True,
        owner_id: typing.Optional[str] = None,
        config: typing.Optional[nio.AsyncClientConfig] = None,
        ssl: bool = True,
        proxy: typing.Optional[str] = None,
        help_command: typing.Optional[typing.Union[Command, typing.Callable[["Context"], typing.Any]]] = None,
        global_message_type: typing.Literal["m.text", "m.notice"] = "m.notice",
        ignore_old_events: bool = True,
        auto_join_rooms: bool = True,
        auto_read_messages: bool = True,
        max_message_cache: int = 1000,
        ignore_self: bool = True,
        import_keys: typing.Tuple[os.PathLike, typing.Optional[str]] = None,
    ):
        if user_id == owner_id and ignore_self is True:
            warnings.warn(
                UserWarning(
                    "User ID and owner ID are the same, but ignore_self is True, meaning no owner systems can be used."
                    " This is probably not what you want."
                )
            )
        try:
            self.loop = asyncio.get_running_loop()
        except RuntimeError:
            self.loop = None
        self.log = logging.getLogger(__name__)
        if store_path:
            if not os.path.exists(store_path):
                self.log.warning("Store path %s does not exist, creating...", store_path)
                os.makedirs(store_path, mode=0o751, exist_ok=True)
            elif not os.path.isdir(store_path):
                raise FileNotFoundError("Store path %s is not a directory!" % store_path)

        if ENCRYPTION_ENABLED:
            if not config:
                config = nio.AsyncClientConfig(encryption_enabled=True, store_sync_tokens=True)
                self.log.info("Encryption support enabled automatically.")
        else:
            self.log.info("Encryption support is not available (are the e2ee extras installed?)")

        super().__init__(
            homeserver,
            user_id,
            device_id,
            store_path=store_path,
            config=config,
            ssl=ssl,
            proxy=proxy,
        )
        self.user_id = user_id
        self.device_id = device_id
        self.store_path = store_path
        self.case_insensitive = case_insensitive
        self.owner_id = owner_id
        self.ignore_self = ignore_self

        if not isinstance(command_prefix, (str, re.Pattern)):
            try:
                iter(command_prefix)
            except TypeError:
                raise TypeError("Command prefix must be a string,  or a regex pattern.") from None
            else:
                self.command_prefix: typing.Tuple[str, ...] = tuple(command_prefix)
        elif isinstance(command_prefix, re.Pattern):
            self.command_prefix: re.Pattern = command_prefix
        else:
            self.command_prefix: typing.Tuple[str] = (command_prefix,)

        if isinstance(self.command_prefix, re.Pattern):
            if self.command_prefix.match("/"):
                self.log.warning(
                    "The prefix '/' may interfere with client-side commands on some clients, such as Element."
                )
        else:
            if "/" in self.command_prefix:
                self.log.warning(
                    "The prefix '/' may interfere with client-side commands on some clients, such as Element."
                )
            if re.search(r"\s", ";".join(command_prefix)):
                raise RuntimeError("Command prefix cannot contain whitespace.")

        self.start_time: typing.Optional[float] = None
        help_cmd = Command(
            "help", DefaultHelpCommand().respond, aliases=["h"], description="Shows a list of commands for this bot"
        )
        if help_command:
            cmd = help_command
            if isinstance(cmd, Command):
                help_cmd = cmd
            elif asyncio.iscoroutinefunction(cmd) or inspect.isfunction(cmd):
                self.log.warning(
                    "Manually changing default help command callback to %r. Please consider passing your own"
                    " Command instance instead."
                )
                help_cmd.callback = cmd
            else:
                raise TypeError("help_command must be a Command instance or a coroutine/function.")
        self._commands = {}
        self._modules = {}
        self._events = {}
        self._event_tasks = []

        self.global_message_type = global_message_type
        self.ignore_old_events = ignore_old_events
        self.auto_join_rooms = auto_join_rooms
        self.auto_read_messages = auto_read_messages

        self.add_event_callback(self.process_message, nio.RoomMessage)  # type: ignore
        self.direct_rooms: dict[str, nio.MatrixRoom] = {}

        self.message_cache: typing.Deque[typing.Tuple[nio.MatrixRoom, nio.RoomMessage]] = deque(
            maxlen=max_message_cache
        )
        self.is_ready = asyncio.Event()
        self._waiting_events = {}

        if self.auto_join_rooms:
            self.log.info("Auto-joining rooms enabled.")
            self.add_event_callback(self._auto_join_room_backlog_callback, nio.InviteMemberEvent)  # type: ignore

        if self.auto_read_messages:
            self.log.info("Auto-updating read receipts enabled.")
            self.add_event_callback(self.update_read_receipts, nio.RoomMessage)

        if import_keys:
            keys_path, keys_password = import_keys
            if not keys_password:
                if sys.stdin.isatty():
                    keys_password = getpass.getpass(f"Password for key import ({keys_path}): ")
                else:
                    raise ValueError(
                        "No password was provided for automatic key import and cannot interactively get password."
                    )

            self.__key_import = pathlib.Path(keys_path), keys_password
        else:
            self.__key_import = None

        self.server_info: typing.Optional[dict] = None
        self.add_command(help_cmd)

    @property
    def supported_server_versions(self) -> typing.List[typing.Tuple[int, int, int]]:
        """
        Returns the supported server versions as a list of `major`, `minor`, `patch` tuples.

        The only time `patch` is >0 is when the server is using a deprecated `r` release.
        All stable releases (`v1`) will have `patch` as 0.
        """
        parsed = []
        if self.server_info:
            for version in self.server_info.get("versions", []):
                if version.startswith("r"):
                    major, minor, patch = map(int, version[1:].split("."))
                else:
                    major, minor = map(int, version[1:].split("."))
                    patch = 0
                parsed.append((major, minor, patch))
        return parsed or [(1, 1, 0)]  # default to 1.1.0 if no server info is available

    def server_supports(self, version: typing.Union[typing.Tuple[int, int], typing.Tuple[int, int, int]]) -> bool:
        """Checks that the server supports at least this matrix version."""
        return any((v >= version for v in self.supported_server_versions))
        # e.g. (1, 1, 0) >= (1, 12, 0) = False | (1, 12, 0) >= (1, 11, 0) = True

    async def mxc_to_http(
        self,
        mxc: str,
        homeserver: Optional[str] = None,
    ) -> Optional[str]:
        """
        Converts an `mxc://` URI to a downloadable HTTP URL.

        This function is identical the [nio.AsyncClient.mxc_to_http()][nio.AsyncClient.mxc_to_http] function,
        however supports matrix 1.10 and below's unauthenticated media automatically.

        :param mxc: The mxc URI
        :param homeserver: The homeserver to download this through (defaults to the bot's homeserver)
        :return: an MXC URL, if applicable
        """
        http: Optional[str] = await super().mxc_to_http(mxc, homeserver)
        if http is not None and not self.server_supports((1, 11, 0)):  # 1.10 and below
            http = http.replace("/_matrix/client/v1/media/download/", "/_matrix/media/r0/download/")
        return http

    async def sync(self, *args, **kwargs) -> U[nio.SyncResponse, nio.SyncError]:
        sync = await super().sync(*args, **kwargs)
        if isinstance(sync, nio.SyncResponse):
            self._populate_dm_rooms(sync)
        return sync

    def _populate_dm_rooms(self, sync: nio.SyncResponse):
        # This function is a historical workaround. It is kept as it is still useful in some cases.
        for room_id, room_info in sync.rooms.join.items():
            for event in room_info.state:
                if isinstance(event, nio.RoomMemberEvent):
                    prev = event.prev_content or {}
                    if event.content.get("is_direct", prev.get("is_direct", False)):
                        self.log.debug("Found DM room in sync: %s", room_id)
                        self.direct_rooms[event.state_key] = self.rooms[room_id]

    async def _auto_join_room_callback(self, room: nio.MatrixRoom, _: nio.InviteMemberEvent):
        """Callback for auto-joining rooms"""
        if self.auto_join_rooms:
            self.log.info("Joining room %s", room.room_id)
            result = await self.join(room.room_id)
            if isinstance(result, nio.JoinError):
                self.log.error("Failed to join room %s: %s", room.room_id, result.message)
            else:
                self.log.info("Joined room %s", room.room_id)

    async def _auto_join_room_backlog_callback(self, room: nio.MatrixRoom, event: nio.InviteMemberEvent):
        """Callback for auto-joining rooms that are backlogged on startup"""
        if event.state_key == self.user_id:
            await self._auto_join_room_callback(room, event)

    @staticmethod
    def latency(event: nio.Event, *, received_at: typing.Optional[float] = None) -> float:
        """Returns the latency for a given event in milliseconds

        :param event: The event to measure latency with
        :param received_at: The optional time the event was received at. If not given, uses the current time.
        :return: The latency in milliseconds"""
        now = received_at or time.time()
        return (now - event.server_timestamp / 1000) * 1000

    def dispatch(self, event_name: typing.Union[str, nio.Event], *args, **kwargs):
        """Dispatches an event to listeners"""
        if event_name in self._events:
            for handler in self._events[event_name]:
                self.log.debug("Dispatching %s to %r" % (event_name, handler))
                try:
                    task = asyncio.create_task(
                        handler(*args, **kwargs), name="DISPATCH_%s_%s" % (handler.__qualname__, os.urandom(3).hex())
                    )
                    self._event_tasks.append(task)
                    task.add_done_callback(
                        lambda *_, **__: self._event_tasks.remove(task) if task in self._event_tasks else None
                    )
                except Exception as e:
                    self.log.exception("Error dispatching %s to %r", event_name, handler, exc_info=e)
        else:
            self.log.debug("%r is not in registered events: %s", event_name, self._events)

    def is_old(self, event: nio.Event) -> bool:
        """Checks if an event was sent before the bot started. Always returns False when ignore_old_events is False"""
        if not self.start_time:
            self.log.warning("have not started yet, using relative age comparison")
            start_time = time.time() - 30  # relative
        else:
            start_time = self.start_time
        if self.ignore_old_events is False:
            return False
        return start_time - event.server_timestamp / 1000 > 0

    async def update_read_receipts(self, room: U[str, nio.MatrixRoom], event: nio.Event):
        """
        Moves the read indicator to the given event in the room.

        !!! info "This is automatically done for you."
            Whenever a message is received, this is automatically called for you.
            As such, your read receipt will always be the most recent message. You rarely need to call this function.

        :param room: The room to update the read receipt in.
        :param event: The event to move the read receipt to.
        :return: Nothing
        """
        room = self._get_id(room)
        if self.is_old(event):
            self.log.debug("Ignoring event %s, sent before bot started.", event.event_id)
            return
        event_id = event.event_id
        result = await self.room_read_markers(room, event_id, event_id)
        if not isinstance(result, nio.RoomReadMarkersResponse):
            msg = result.message if isinstance(result, nio.ErrorResponse) else "?"
            self.log.warning("Failed to update read receipts for %s: %s", room, msg)
        else:
            self.log.debug("Updated read receipts for %s to %s.", room, event_id)

    async def process_message(self, room: nio.MatrixRoom, event: nio.RoomMessage) -> None:
        """Processes a message and runs the command it is trying to invoke if any."""
        if self.start_time is None:
            raise RuntimeError("Bot has not started yet!")

        self.message_cache.append((room, event))
        self.dispatch("message", room, event)
        if not isinstance(event, nio.RoomMessageText):
            self.log.debug("Ignoring non-text message %r", event)
            return
        if event.sender == self.user and self.ignore_self is True:
            self.log.debug("Ignoring message sent by self.")
            return
        if self.is_old(event):
            age = self.start_time - event.server_timestamp / 1000
            self.log.debug("Ignoring message sent {:.0f} seconds before startup.".format(age))
            return

        if self.case_insensitive:
            content = event.body.casefold()
        else:
            content = event.body

        def get_prefix(c: str) -> typing.Union[str, None]:
            if isinstance(self.command_prefix, re.Pattern):
                _m = re.match(self.command_prefix, c)
                if _m:
                    return _m.group(0)
            else:
                for pfx in self.command_prefix:
                    if c.startswith(pfx):
                        return pfx

        matched_prefix = get_prefix(content)
        if matched_prefix:
            try:
                command = original_command = content[len(matched_prefix) :].splitlines()[0].split(" ")[0]
            except IndexError:
                self.log.info(
                    "Failed to parse message %r - message terminated early (was the content *just* the prefix?)",
                    event.body,
                )
                return
            command = self.get_command(command)
            if command:
                if command.disabled is True:
                    error = CommandDisabledError(command)
                    self.dispatch("command_error", command, error)
                    return

                context = command.construct_context(
                    self,
                    room=room,
                    src_event=event,
                    invoking_prefix=matched_prefix,
                    meta=matched_prefix + original_command,
                )

                try:
                    if not await context.command.can_run(context):
                        raise CheckFailure(None, "Unknown check failure")
                except CheckFailure as err:
                    self.dispatch("command_error", context, err)
                    raise

                def _task_callback(t: asyncio.Task):
                    try:
                        exc = t.exception()
                    except asyncio.CancelledError:
                        self.dispatch("command_cancelled", context, t)
                    else:
                        if exc:
                            self.dispatch("command_error", context, CommandError(exception=exc))
                        else:
                            self.dispatch("command_complete", context, t)
                    if hasattr(context, "_perf_timer"):
                        self.log.debug(
                            "Command %r finished in %.2f seconds",
                            command.name,
                            time.perf_counter() - context._perf_timer,
                        )

                self.log.debug(f"Running command {command.name} with context {context!r}")
                try:
                    task = asyncio.create_task(await command.invoke(context))
                    context._task = task
                    context._perf_timer = time.perf_counter()
                except CommandArgumentsError as e:
                    self.dispatch("command_error", context, e)
                except Exception as e:
                    self.log.exception("Failed to invoke command %s", command.name)
                    self.dispatch("command_error", context, CommandError(exception=e))
                else:
                    task.add_done_callback(_task_callback)
            else:
                self.log.debug(f"Command {original_command!r} not found.")

    def is_owner(self, user_id: str) -> bool:
        """
        Checks whether a user is the owner of the bot.

        :param user_id: The user ID to check.
        :return: Whether the user is the owner.
        """
        if not self.owner_id:
            self.log.warning("Attempted to check for owner, but no owner ID was set!")
            return False  # no owner ID set.
        return self.owner_id == user_id

    def mount_module(self, import_path: str) -> typing.Optional[list[Command]]:
        """Mounts a module including all of its commands.

        Must be a subclass of niobot.commands.Module, or else this function will not work.

        ??? danger "There may not be an event loop running when this function is called."
            If you are calling this function before you call `bot.run()`, it is entirely possible that you don't have
            a running [asyncio][] event loop. If you use the event loop in `Module.__init__`, you will get an error,
            and the module will fail the mount.

            You can get around this by deferring mounting your modules until the `bot.on_ready` event is fired,
            at which point not only will the first full sync have completed (meaning the bot has all of its caches
            populated), but the event loop will be running.

        :param import_path: The import path (such as modules.file), which would be ./modules/file.py in a file tree.
        :returns: Optional[list[Command]] - A list of commands mounted. None if the module's setup() was called.
        :raise ImportError: The module path is incorrect of there was another error while importing
        :raise TypeError: The module was not a subclass of Module.
        :raise ValueError: There was an error registering a command (e.g. name conflict)
        """
        added = []
        module = importlib.import_module(import_path)
        if hasattr(module, "setup") and callable(module.setup):
            # call setup
            self.log.debug("Calling module-defined setup() function rather than doing it manually.")
            _c = self.commands.copy()
            _e = self._events.copy()
            module.setup(self)
            if _c == self.commands and _e == self._events:
                self.log.warning(
                    "Module %r did not add any commands or events with the custom setup function?",
                    module,
                )
            return

        self.log.debug("%r does not have its own setup() - auto-discovering commands and events", module)
        for _, item in inspect.getmembers(module):
            if inspect.isclass(item):
                if getattr(item, "__is_nio_module__", False):
                    if item in self._modules:
                        raise ValueError("%r is already loaded." % item.__class__.__name__)
                    instance = item(self)
                    if not isinstance(instance, Module):
                        raise TypeError("%r is not a subclass of Module." % instance.__class__.__name__)
                    instance.__setup__()
                    self._modules[item] = instance
                    added += list(instance.list_commands())
                    self.log.debug(
                        "Loaded %d commands from %r",
                        len(set(instance.list_commands())),
                        instance.__class__.__qualname__,
                    )
                else:
                    self.log.debug("%r does not appear to be a niobot module", item)
        return added

    def unmount_module(self, module: Module) -> None:
        """
        Does the opposite of mounting the module.
        This will remove any commands that have been added to the bot from the given module.

        :param module: The module to unmount
        """
        self.log.debug("Unmounting module %r", module)
        module.__teardown__()

    @property
    def commands(self) -> dict[str, Command]:
        """Returns the internal command register.

        !!! warning
            Modifying any values here will update the internal register too.

        !!! note
            Aliases of commands are treated as their own command instance. You will see the same command show up as a
            value multiple times if it has aliases.

            You can check if two commands are identical by comparing them (`command1instance == command2instance`)
        """
        return self._commands

    @property
    def modules(self) -> dict[typing.Type, Module]:
        """Returns the internal module register.

        !!! warning
            Modifying any values here will update the internal register too.
        """
        return self._modules

    def get_command(self, name: str) -> typing.Optional[Command]:
        """Attempts to retrieve an internal command

        :param name: The name of the command to retrieve
        :return: The command, if found. None otherwise.
        """
        return self._commands.get(name)

    def add_command(self, command: Command) -> None:
        """Adds a command to the internal register

        if a name or alias is already registered, this throws a ValueError.
        Otherwise, it returns None."""
        if self.get_command(command.name):
            raise ValueError(f"Command or alias {command.name} is already registered.")
        if any((self.get_command(alias) for alias in command.aliases)):
            raise ValueError(f"Command or alias for {command.name} is already registered.")

        self._commands[command.name] = command
        self.log.debug("Registered command %r into %s", command, command.name)
        for alias in command.aliases:
            self._commands[alias] = command
            self.log.debug("Registered command %r into %s", command, alias)

    def remove_command(self, command: Command) -> None:
        """Removes a command from the internal register.

        If the command is not registered, this is a no-op."""
        if not self.get_command(command.name):
            return

        self.log.debug("Removed command %r from the register.", self._commands.pop(command.name, None))
        for alias in command.aliases:
            self.log.debug("Removed command %r from the register.", self._commands.pop(alias, None))

    def command(self, name: typing.Optional[str] = None, **kwargs):
        """Registers a command with the bot."""
        cls = kwargs.pop("cls", Command)

        def decorator(func):
            nonlocal name
            name = name or func.__name__
            command = cls(name, func, **kwargs)
            self.add_command(command)
            return func

        return decorator

    def add_event_listener(self, event_type: typing.Union[str, nio.Event], func):
        self._events.setdefault(event_type, [])
        self._events[event_type].append(func)
        self.log.debug("Added event listener %r for %r", func, event_type)

    def on_event(self, event_type: typing.Optional[typing.Union[str, nio.Event]] = None):
        """Wrapper that allows you to register an event handler.

        Event handlers **must** be async.

        if event_type is None, the function name is used as the event type.

        Please note that if you pass a [Event][nio.events.room_events.Event], you are responsible for capturing errors.
        """

        def wrapper(func):
            nonlocal event_type
            event_type = event_type or func.__name__
            if isinstance(event_type, str):
                if event_type.startswith("on_"):
                    self.log.warning("No events start with 'on_' - stripping prefix from %r", event_type)
                    event_type = event_type[3:]
            self.add_event_listener(event_type, func)
            return func

        return wrapper

    def remove_event_listener(self, function):
        for event_type, functions in self._events.items():
            if function in functions:
                self._events[event_type].remove(function)
                self.log.debug("Removed %r from event %r", function, event_type)

    async def set_room_nickname(
        self,
        room: U[str, nio.MatrixRoom],
        new_nickname: str = None,
        user: typing.Optional[U[str, nio.MatrixUser]] = None,
    ) -> nio.RoomPutStateResponse:
        """
        Changes the user's nickname in the given room.

        :param room: The room to change the nickname in.
        :param new_nickname: The new nickname. If None, defaults to the user's display name.
        :param user: The user to update. Defaults to the bot's user.
        :return: The response from the server.
        :raise: GenericMatrixError - The request failed.
        """
        room_id = self._get_id(room)
        user = user or self.user_id
        user_id = self._get_id(user)
        profile = await self.get_profile(user_id)
        if isinstance(profile, nio.ProfileGetError):
            raise GenericMatrixError("Failed to get profile", response=profile)

        result = await self.room_put_state(
            room_id,
            "m.room.member",
            {"membership": "join", "displayname": new_nickname, "avatar_url": profile.avatar_url},
            user_id,
        )
        if isinstance(result, nio.RoomPutStateError):
            raise GenericMatrixError("Failed to set nickname", response=result)
        return result

    async def room_send(
        self,
        room_id: str,
        message_type: str,
        content: dict,
        tx_id: typing.Optional[str] = None,
        ignore_unverified_devices: bool = True,
    ) -> U[nio.RoomSendResponse, nio.RoomSendError]:
        return await super().room_send(
            room_id,
            message_type,
            content,
            tx_id,
            ignore_unverified_devices,
        )

    def get_cached_message(self, event_id: str) -> typing.Optional[typing.Tuple[nio.MatrixRoom, nio.RoomMessage]]:
        """Fetches a message from the cache.

        This returns both the room the message was sent in, and the event itself.

        If the message is not in the cache, this returns None."""
        for room, event in self.message_cache:
            if event_id == event.event_id:
                return room, event

    async def fetch_message(self, room_id: str, event_id: str):
        """Fetches a message from the server."""
        cached = self.get_cached_message(event_id)
        if cached:
            return cached

        result: typing.Union[nio.RoomGetEventError, nio.RoomGetEventResponse]
        result = await self.room_get_event(room_id, event_id)
        if isinstance(result, nio.RoomGetEventError):
            raise NioBotException(f"Failed to fetch message {event_id} from {room_id}: {result}", original=result)
        return result

    async def wait_for_message(
        self,
        room_id: typing.Optional[str] = None,
        sender: typing.Optional[str] = None,
        check: typing.Optional[typing.Callable[[nio.MatrixRoom, nio.RoomMessageText], typing.Any]] = None,
        *,
        timeout: typing.Optional[float] = None,
        msg_type: typing.Type[nio.RoomMessage] = nio.RoomMessageText,
    ) -> typing.Optional[typing.Tuple[nio.MatrixRoom, nio.RoomMessage]]:
        """Waits for a message, optionally with a filter.

        If this function times out, asyncio.TimeoutError is raised.

        :param room_id: The room ID to wait for a message in. If None, waits for any room.
        :param sender: The user ID to wait for a message from. If None, waits for any sender.
        :param check: A function to check the message with. If the function returns False, the message is ignored.
        :param timeout: The maximum time to wait for a message. If None, waits indefinitely.
        :param msg_type: The type of message to wait for. Defaults to nio.RoomMessageText.
        :return: The room and message that was received.
        """
        event = asyncio.Event()
        value = None

        async def event_handler(_room, _event):
            if not isinstance(_event, msg_type):
                self.log.debug("Ignoring non-text message %r", _event)
            if room_id and _room.room_id != room_id:
                self.log.debug("Ignoring bubbling message from %r (vs %r)", _room.room_id, room_id)
                return False
            if sender and _event.sender != sender:
                self.log.debug("Ignoring bubbling message from %r (vs %r)", _event.sender, sender)
                return False
            if check:
                try:
                    result = await force_await(check, _room, _event)
                except Exception as e:
                    self.log.error("Error in check function: %r", e, exc_info=e)
                    return False
                if not result:
                    self.log.debug("Ignoring bubbling message, check was false")
                    return False
            event.set()
            nonlocal value
            value = _room, _event

        self.on_event("message")(event_handler)
        try:
            await asyncio.wait_for(event.wait(), timeout=timeout)
        finally:
            self.remove_event_listener(event_handler)
        return value

    @staticmethod
    async def markdown_to_html(text: str) -> str:
        """
        Converts markdown to HTML.

        :param text: The markdown to render as HTML
        :return: the rendered HTML
        """
        parsed = await run_blocking(marko.parse, text)
        if parsed.children:
            rendered = await run_blocking(marko.render, parsed)
        else:
            rendered = text
        return rendered

    @property
    @deprecated("niobot.NioBot.markdown_to_html")
    def _markdown_to_html(self) -> typing.Callable[[str], typing.Awaitable[str]]:
        return self.markdown_to_html

    @staticmethod
    def _get_id(obj: typing.Union[nio.Event, nio.MatrixRoom, nio.MatrixUser, str, typing.Any]) -> str:
        """
        Gets the id of most objects as a string.

        :param obj: The object whose ID to get, or the ID itself.
        :type obj: typing.Union[nio.Event, nio.MatrixRoom, nio.MatrixUser, str, Any]
        :returns: the ID of the object
        :raises: ValueError - the Object doesn't have an ID
        """
        if hasattr(obj, "event_id"):
            return obj.event_id
        if hasattr(obj, "room_id"):
            return obj.room_id
        if hasattr(obj, "user_id"):
            return obj.user_id
        if isinstance(obj, str):
            return obj
        raise ValueError("Unable to determine ID of object: %r" % obj)

    @staticmethod
    def generate_mx_reply(room: nio.MatrixRoom, event: nio.RoomMessageText) -> str:
        """Generates a reply string for a given event."""
        return (
            "<mx-reply>"
            "<blockquote>"
            '<a href="{reply_url}">{reply}</a> '
            '<a href="{user_url}">{user}</a><br/>'
            "</blockquote>"
            "</mx-reply>".format(
                reply_url="https://matrix.to/#/{}:{}/{}".format(
                    room.room_id, room.machine_name.split(":")[1], event.event_id
                ),
                reply=event.body,
                user_url="https://matrix.to/#/{}".format(event.sender),
                user=event.sender,
            )
        )

    async def _recursively_upload_attachments(
        self, base: "BaseAttachment", encrypted: bool = False, __previous: typing.Optional[list] = None
    ) -> list[typing.Union[nio.UploadResponse, nio.UploadError, None]]:
        """Recursively uploads attachments."""
        previous = (__previous or []).copy()
        if not base.url:
            self.log.info("Uploading attachment %r (encrypted: %r)", base, encrypted)
            previous.append(await base.upload(self, encrypted))
        if hasattr(base, "thumbnail") and base.thumbnail and not base.url:
            self.log.info("Uploading thumbnail %r (encrypted: %r)", base.thumbnail, encrypted)
            previous += await self._recursively_upload_attachments(base.thumbnail, encrypted, previous)
        return previous

    @typing.overload
    async def get_dm_rooms(self) -> typing.Dict[str, typing.List[str]]:
        """
        Gets all DM rooms stored in account data.

        :return: A dictionary containing user IDs as keys, and lists of room IDs as values.
        """
        ...

    @typing.overload
    async def get_dm_rooms(self, user: U[nio.MatrixUser, str]) -> typing.List[str]:
        """
        Gets DM rooms for a specific user.

        :param user: The user to fetch DM rooms for.
        :return: A list of room IDs
        """
        ...

    async def get_dm_rooms(
        self,
        user: typing.Optional[U[nio.MatrixUser, str]] = None,
    ) -> typing.Union[typing.Dict[str, typing.List[str]], typing.List[str]]:
        """
        Gets DM rooms, optionally for a specific user.

        If no user is given, this returns a dictionary of user IDs to lists of rooms.

        :param user: The user ID or object to get DM rooms for.
        :return: A dictionary of user IDs to lists of rooms, or a list of rooms.
        """
        result = await self.list_direct_rooms()
        if isinstance(result, nio.DirectRoomsErrorResponse):
            if result.status_code == "M_NOT_FOUND":
                # No DM rooms for this account are known
                return {} if user is None else []
            raise GenericMatrixError("Failed to get DM rooms", response=result)
        if user:
            user_id = self._get_id(user)
            return result.rooms.get(user_id, [])
        return result.rooms

    async def create_dm_room(
        self,
        user: U[nio.MatrixUser, str],
    ) -> nio.RoomCreateResponse:
        """
        Creates a DM room with a given user.

        :param user: The user to create a DM room with.
        :return: The response from the server.
        """
        user_id = self._get_id(user)
        result = await self.room_create(
            is_direct=True,
            invite=[user_id],
            visibility=nio.api.RoomVisibility.private,
            preset=nio.api.RoomPreset.trusted_private_chat,
        )
        if isinstance(result, nio.RoomCreateError):
            raise GenericMatrixError("Failed to create DM room", response=result)

        dm_rooms = {user_id: [result.room_id]}

        logging.debug(f"create_dm_user: user_id {user_id} self.user_id {self.user_id}")
        # Trying to send m.direct type eventfrom nio.responses import DirectRoomsErrorResponse, DirectRoomsResponse
        logging.debug(f"create_dm_user: setting m.direct type with rooms {nio.Api.to_json(dm_rooms)}")
        await self._send(
            nio.DirectRoomsResponse,
            "PUT",
            nio.Api._build_path(
                ["user", self.user_id, "account_data", "m.direct"],
                {"access_token": self.access_token},
            ),
            nio.Api.to_json(dm_rooms),
        )

        return result

    async def send_message(
        self,
        room: U[nio.MatrixRoom, nio.MatrixUser, str],
        content: typing.Optional[str] = None,
        file: typing.Optional[BaseAttachment] = None,
        reply_to: typing.Optional[U[nio.RoomMessage, str]] = None,
        message_type: typing.Optional[str] = None,
        *,
        content_type: typing.Literal["plain", "markdown", "html", "html.raw"] = "markdown",
        override: typing.Optional[dict] = None,
        mentions: typing.Optional[Mentions] = None,
    ) -> nio.RoomSendResponse:
        """
        Sends a message. Doesn't get any more simple than this.

        ??? tip "DMs"
            As of v1.1.0, you can now send messages to users (either a [nio.MatrixUser][nio.rooms.MatrixUser]
            or a user ID string),
            and a direct message room will automatically be created for you if one does not exist, using an existing
            one if it does.

        !!! tip "Content Type"
            **Separate to `message_type`**, `content_type` controls what sort of parsing and formatting will be applied
            to the provided content. This is useful for sending messages that are not markdown, or for sending HTML.
            Before, all content was assumed to be markdown, and was parsed as such. However, this may cause
            undesirable effects if you are sending messages that are not markdown.

            * **`plain`** - No parsing or formatting is applied, and the content is sent as-is.
            * **`markdown`** - The content is parsed as markdown and rendered as HTML, with a fallback plain text
            body. This is the default.
            * **`html`** - The content is sent as HTML, with no fallback to plain text. If BeautifulSoup is installed,
            the provided content will be sanitised and pretty-printed before sending.
            ** **`html.raw`** - The content is sent as HTML, with no fallback to plain text,
            nor sanitising or formatting.

        :param room: The room or to send this message to
        :param content: The content to send. Cannot be used with file.
        :param file: A file to send, if any. Cannot be used with content.
        :param reply_to: A message to reply to.
        :param message_type: The message type to send. If none, defaults to NioBot.global_message_type,
        which itself is `m.notice` by default.
        :param override: A dictionary containing additional properties to pass to the body.
        Overrides existing properties.
        :param content_type: The type of content to send. Defaults to "markdown".
        :param mentions: Intentional mentions to send with the message.
        :return: The response from the server.
        :raises MessageException: If the message fails to send, or if the file fails to upload.
        :raises ValueError: You specified neither file nor content.
        :raises RuntimeError: An internal error occured. A room was created, but is not in the bot room list.
        """
        if file and BaseAttachment is None:
            raise ValueError("You are missing required libraries to use attachments.")
        if not any((content, file)):
            raise ValueError("You must specify either content or file.")

        if isinstance(room, nio.MatrixUser) or (isinstance(room, str) and room.startswith("@")):
            _user = room
            room = None
            rooms = await self.get_dm_rooms(_user)
            logging.debug(f"send_message get_dm_rooms returns rooms {json.dumps(rooms)}")

            if rooms:
                for r_id in rooms:
                    if r_id in self.rooms:
                        room = r_id
                        break
                    else:
                        logging.warning(f"room {r_id} not found in bot.rooms")
            if not room:
                logging.info("creating dm room for user {_user}")
                response = await self.create_dm_room(_user)

                room = self.rooms.get(response.room_id)
                if not room:
                    raise RuntimeError(
                        "DM room %r was created, but could not be found in the room list!" % response.room_id
                    )

        self.log.debug("Send message resolved room to %r", room)

        body: dict[str, typing.Any] = {
            "msgtype": message_type or self.global_message_type,
        }

        if file is not None:
            self.log.info("Recursively uploading %r.", file)
            # We need to upload the file first.
            responses = await self._recursively_upload_attachments(file, encrypted=getattr(file, "encrypted", False))
            if any((isinstance(response, nio.UploadError) for response in responses)):
                raise MessageException(
                    "Failed to upload media.", tuple(filter(lambda x: isinstance(x, nio.UploadError), responses))[0]
                )

            body = file.as_body(content)
        else:
            body["body"] = content
            if content_type == "markdown":
                parsed = await run_blocking(marko.parse, content)
                if parsed.children:
                    rendered = await run_blocking(marko.render, parsed)
                    body["formatted_body"] = rendered
                    body["format"] = "org.matrix.custom.html"
            elif content_type == "html":
                if BeautifulSoup:
                    soup = BeautifulSoup(content)
                    content_new = soup.prettify("utf-8", "minimal")
                else:
                    self.log.debug("Content type was HTML, however BeautifulSoup is not installed. Treating as raw.")
                    content_new = content
                body["formatted_body"] = content_new
                body["format"] = "org.matrix.custom.html"
            elif content_type == "html.raw":
                body["formatted_body"] = content
                body["format"] = "org.matrix.custom.html"
        if reply_to and isinstance(reply_to, nio.RoomMessageText) and isinstance(room, nio.MatrixRoom):
            body["formatted_body"] = "{}{}".format(
                self.generate_mx_reply(room, reply_to), body.get("formatted_body", body["body"])
            )
            body["format"] = "org.matrix.custom.html"

        if reply_to:
            body["m.relates_to"] = {"m.in_reply_to": {"event_id": self._get_id(reply_to)}}
        if mentions:
            body.update(mentions.as_body())

        if override:
            body.update(override)
        try:
            async with Typing(self, self._get_id(room)):
                response = await self.room_send(
                    self._get_id(room),
                    "m.room.message",
                    body,
                )
        except RuntimeError:  # already typing
            response = await self.room_send(
                self._get_id(room),
                "m.room.message",
                body,
            )
        if isinstance(response, nio.RoomSendError):
            raise MessageException("Failed to send message.", response)
        return response

    async def edit_message(
        self,
        room: U[nio.MatrixRoom, str],
        message: U[nio.Event, str],
        content: str,
        *,
        message_type: typing.Optional[str] = None,
        content_type: typing.Literal["plain", "markdown", "html", "html.raw"] = "markdown",
        mentions: typing.Optional[Mentions] = None,
        override: typing.Optional[dict] = None,
    ) -> nio.RoomSendResponse:
        """
        Edit an existing message. You must be the sender of the message.

        You also cannot edit messages that are attachments.

        :param room: The room the message is in.
        :param message: The message to edit.
        :param content: The new content of the message.
        :param message_type: The new type of the message (i.e. m.text, m.notice. Defaults to client.global_message_type)
        :param override: A dictionary containing additional properties to pass to the body.
        Overrides existing properties.
        :param content_type: The type of content to send. Defaults to "markdown".
        :raises RuntimeError: If you are not the sender of the message.
        :raises TypeError: If the message is not text.
        """

        room = self._get_id(room)

        event_id = self._get_id(message)
        message_type = message_type or self.global_message_type
        content_dict = {
            "msgtype": message_type,
            "body": content,
            "format": "org.matrix.custom.html",
            "formatted_body": await self.markdown_to_html(content),
        }

        body = {
            "msgtype": message_type,
            "m.new_content": {**content_dict},
            "m.relates_to": {
                "rel_type": "m.replace",
                "event_id": event_id,
            },
            "body": "* %s" % content_dict["body"],
        }

        if content_type == "markdown":
            parsed = await run_blocking(marko.parse, content)
            if parsed.children:
                rendered = await run_blocking(marko.render, parsed)
                body["formatted_body"] = rendered
                body["format"] = "org.matrix.custom.html"
        elif content_type == "html":
            if BeautifulSoup is not None:
                soup = BeautifulSoup(content)
                content_new = soup.prettify("utf-8", "minimal")
            else:
                self.log.debug("Content type was HTML, however BeautifulSoup is not installed. Treating as raw.")
                content_new = content
            body["formatted_body"] = content_new
            body["format"] = "org.matrix.custom.html"
        elif content_type == "html.raw":
            body["formatted_body"] = content
            body["format"] = "org.matrix.custom.html"
        if mentions:
            body.update(mentions.as_body())
        if override:
            body.update(override)
        async with Typing(self, room):
            response = await self.room_send(
                room,
                "m.room.message",
                body,
            )
        if isinstance(response, nio.RoomSendError):
            raise MessageException("Failed to edit message.", response)
        # Forcefully clear typing
        await self.room_typing(room, False)
        return response

    async def delete_message(
        self, room: U[nio.MatrixRoom, str], message_id: U[nio.RoomMessage, str], reason: typing.Optional[str] = None
    ) -> nio.RoomRedactResponse:
        """
        Delete an existing message. You must be the sender of the message.

        :param room: The room the message is in.
        :param message_id: The message to delete.
        :param reason: The reason for deleting the message.
        :raises RuntimeError: If you are not the sender of the message.
        :raises MessageException: If the message fails to delete.
        """
        room = self._get_id(room)
        message_id = self._get_id(message_id)
        response = await self.room_redact(room, message_id, reason=reason)
        if isinstance(response, nio.RoomRedactError):
            raise MessageException("Failed to delete message.", response)
        return response

    async def add_reaction(
        self, room: U[nio.MatrixRoom, str], message: U[nio.RoomMessage, str], emoji: str
    ) -> nio.RoomSendResponse:
        """
        Adds an emoji "reaction" to a message.

        :param room: The room the message is in.
        :param message: The event ID or message object to react to.
        :param emoji: The emoji to react with (e.g. `\N{CROSS MARK}` = ❌)
        :return: The response from the server.
        :raises MessageException: If the message fails to react.
        """
        body = {
            "m.relates_to": {
                "event_id": self._get_id(message),
                "rel_type": "m.annotation",
                "key": emoji,
            }
        }
        response = await self.room_send(
            self._get_id(room),
            "m.reaction",
            body,
        )
        if isinstance(response, nio.RoomSendError):
            raise MessageException("Failed to add reaction.", response)
        return response

    async def redact_reaction(self, room: U[nio.MatrixRoom, str], reaction: U[nio.RoomSendResponse, str]):
        """Alias for NioBot.delete_message, but more appropriately named for reactions."""
        response = await self.room_redact(
            self._get_id(room),
            self._get_id(reaction),
        )
        if isinstance(response, nio.RoomRedactError):
            raise MessageException("Failed to delete reaction.", response)
        return response

    async def start(
        self,
        password: typing.Optional[str] = None,
        access_token: typing.Optional[str] = None,
        sso_token: typing.Optional[str] = None,
    ) -> None:
        """Starts the bot, running the sync loop."""
        self.loop = asyncio.get_event_loop()
        self.dispatch("event_loop_ready")
        if self.__key_import:
            self.log.info("Starting automatic key import")
            await self.import_keys(*map(str, self.__key_import))

        if password or sso_token:
            if password:
                self.log.critical(
                    "Logging in with a password is insecure, slow, and clunky. "
                    "An access token will be issued after logging in, please use that. For more information, see:"
                    "https://docs.nio-bot.dev/guides/001-getting-started/#why-is-logging-in-with-a-password-so-bad"
                )
            self.log.info("Logging in with a password or SSO token")
            login_response = await self.login(password=password, token=sso_token, device_name=self.device_id)
            if isinstance(login_response, nio.LoginError):
                raise LoginException("Failed to log in.", login_response)

            self.log.info("Logged in as %s", login_response.user_id)
            self.log.debug("Logged in: {0.access_token}, {0.user_id}".format(login_response))
            self.start_time = time.time()
        elif access_token:
            self.log.info("Logging in with existing access token.")
            if self.store_path:
                try:
                    self.load_store()
                except FileNotFoundError:
                    self.log.warning("Failed to load store.")
                except nio.LocalProtocolError as e:
                    self.log.warning("No store? %r", e, exc_info=e)
            self.access_token = access_token
            self.start_time = time.time()
        else:
            raise LoginException("You must specify either a password/SSO token or an access token.")

        if self.should_upload_keys:
            self.log.info("Uploading encryption keys...")
            response = await self.keys_upload()
            if isinstance(response, nio.KeysUploadError):
                self.log.critical("Failed to upload encryption keys. Encryption may not work. Error: %r", response)
            else:
                self.log.info("Uploaded encryption keys.")
        self.log.info("Fetching server details...")
        response = await self.send("GET", "/_matrix/client/versions")
        if response.status != 200:
            self.log.warning("Failed to fetch server details. Status: %d", response.status)
        else:
            self.server_info = await response.json()
            self.log.debug("Server details: %r", self.server_info)
        self.log.info("Performing first sync...")
        result = await self.sync(timeout=30000, full_state=True, set_presence="unavailable")
        if not isinstance(result, nio.SyncResponse):
            raise NioBotException("Failed to perform first sync.", result)
        self.is_ready.set()
        self.dispatch("ready", result)
        self.log.info("Starting sync loop")
        try:
            await self.sync_forever(
                timeout=30000,
                full_state=True,
                set_presence="online",
            )
        finally:
            self.log.info("Closing http session.")
            await self.close()

    def run(
        self,
        *,
        password: typing.Optional[str] = None,
        access_token: typing.Optional[str] = None,
        sso_token: typing.Optional[str] = None,
    ) -> None:
        """
        Runs the bot, blocking the program until the event loop exists.
        This should be the last function to be called in your script, as once it exits, the bot will stop running.

        Note:
            This function is literally just asyncio.run(NioBot.start(...)), so you won't have much control over the
            asyncio event loop. If you want more control, you should use await NioBot.start(...) instead.

        :param password: The password to log in with.
        :param access_token: An existing login token.
        :param sso_token: An SSO token to sign in with.
        :return:
        """
        asyncio.run(self.start(password=password, access_token=access_token, sso_token=sso_token))



"""
A simple, interactive niobot example.

This script will ask for initial information interactively, and then writes it to a config file.
So, make sure you've got an interactive session on first run.

This example was written in nio-bot version 1.1.0b2.
"""

import getpass
from base64 import b16decode, b16encode
from configparser import ConfigParser
from pathlib import Path

import niobot

__author__ = "nexy7574 (@nex:nex7574.co.uk)"
__version__ = "1.1.1"


CONFIG_PATH = Path("./config.ini")
CONFIG_PATH.touch(exist_ok=True)
config = ConfigParser()
config.read(CONFIG_PATH)
if "niobot" not in config:
    homeserver = input("Homeserver URL (e.g. https://matrix.example.org): ")
    username = input("Username (e.g. @username:example.com): ")
    try:
        password = getpass.getpass("Password (CTRL+C to use token instead): ")
        token = None
    except KeyboardInterrupt:
        token = getpass.getpass("Token: ")
        password = None
    device_id = input("Device ID/Name (default: niobot-simple-example): ") or "niobot-simple-example"
    prefix = input("Bot prefix (default: !): ") or "!"
    owner_id = input("Owner ID (e.g. @example:example.com, default: None): ") or None
    config["niobot"] = {
        "homeserver": homeserver,
        "username": username,
        "device_id": device_id,
        "prefix": prefix,
        "owner_id": owner_id,
    }
    if password:
        config["niobot"]["password"] = b16encode(password.encode()).decode()
    if token:
        config["niobot"]["token"] = b16encode(token.encode()).decode()
    with CONFIG_PATH.open("w") as f:
        config.write(f)
else:
    homeserver = config["niobot"]["homeserver"]
    username = config["niobot"]["username"]
    password = config["niobot"].get("password")
    if password:
        password = b16decode(password.encode()).decode()
    token = config["niobot"].get("token")
    if token:
        token = b16decode(token.encode()).decode()
    if not any((password, token)):
        raise ValueError("Either password or token must be set in config")
    device_id = config["niobot"]["device_id"]
    prefix = config["niobot"]["prefix"]
    owner_id = config["niobot"].get("owner_id")

bot = niobot.NioBot(
    homeserver=homeserver,
    user_id=username,
    device_id=device_id,
    store_path=config["niobot"].get("store_path"),
    command_prefix=prefix,
    owner_id=owner_id,
)


@bot.on_event("ready")
async def on_ready(_):
    print(f"Logged in as {bot.user_id}")
    print(f"Prefix: {bot.command_prefix}")
    if config["niobot"].get("password"):
        print("Removing password in config in favour of access token.")
        config["niobot"].pop("password")
        config["niobot"]["token"] = bot.access_token
        with CONFIG_PATH.open("w") as file:
            config.write(file)


@bot.command("ping")
async def ping_command(ctx: niobot.Context):
    """A simple ping command"""
    latency = ctx.latency
    await ctx.respond(f"Pong! `{latency:.2f}ms` latency.")


@bot.command("wave")
async def wave_to(ctx: niobot.Context, user: niobot.MatrixUser):
    """Wave to a user"""
    await ctx.respond(f"*waves to {user.display_name}*")


bot.run(access_token=token, password=password)



import pytest

from niobot.utils import unblocking


async def fake_coro(return_value=1):
    return return_value


def fake_sync(return_value=2):
    return return_value


@pytest.mark.parametrize(
    "function, kwargs, expected, _raise",
    [
        (fake_coro, {"return_value": 5}, 5, TypeError),
        (fake_sync, {"return_value": 6}, 6, None),
        (fake_coro, {}, 1, TypeError),
        (fake_sync, {}, 2, None),
    ],
)
@pytest.mark.asyncio
async def test_run_blocking(function, kwargs, expected, _raise):
    non_coro = unblocking.run_blocking(function, **kwargs)
    if _raise is not None:
        with pytest.raises(_raise):
            await non_coro
    else:
        assert (await non_coro) == expected, "Unexpected result: %r != %r" % (non_coro, expected)


@pytest.mark.parametrize(
    "function, kwargs, expected",
    [
        (fake_coro, {"return_value": 5}, 5),
        (fake_sync, {"return_value": 6}, 6),
        (fake_coro, {}, 1),
        (fake_sync, {}, 2),
    ],
)
@pytest.mark.asyncio
async def test_force_await(function, kwargs, expected):
    non_coro = unblocking.force_await(function, **kwargs)
    assert (await non_coro) == expected, "Unexpected result: %r != %r" % (non_coro, expected)
